import datetime
import json
from enum import Enum
from functools import cached_property
from os import PathLike
from pathlib import Path, PurePath
from typing import (
    Annotated,
    Any,
    Literal,
)

from anyio import Path as AsyncPath
from pydantic import BaseModel, Field

type FilePath = str | PathLike[str]


class CreateChatResponse(BaseModel):
    chat_uuid: str


class RunWorkflowRequest(BaseModel):
    chat_uuid: str
    workflow_id: str


# note: before adding fields here, probably update
# get_workflow_run_by_trigger db query
class PrReviewWorkflowInput(BaseModel):
    repo_owner: str
    repo_name: str
    pr_number: int
    branch: str | None = None
    banner_comment_id: int | None = None
    # PR metadata fields - populated from webhook payload
    pr_title: str | None = None
    pr_url: str | None = None
    pr_author: str | None = None
    pr_author_avatar_url: str | None = None
    pr_additions: int | None = None
    pr_deletions: int | None = None
    pr_changed_files: int | None = None
    head_sha: str | None = None


class SlackWorkflowInput(BaseModel):
    discriminator: Literal["slack_workflow"] = "slack_workflow"
    channel_id: str
    thread_ts: str
    slack_url: str | None = None
    channel_name: str | None = None
    message_ts: str | None = None
    message_text: str | None = None


WorkflowInput = PrReviewWorkflowInput | SlackWorkflowInput


class WorkflowTriggerRequest(BaseModel):
    workflow_name: str
    workflow_input: WorkflowInput


class WorkflowTriggerResponse(BaseModel):
    chat_uuid: str


class ExecutionEndResponse(BaseModel):
    execution_ended: bool


class SignalType(str, Enum):
    disconnect = "disconnect"

    def __str__(self) -> str:
        return self.value


class GitInfo(BaseModel):
    branch: str
    remote: str | None


class PythonEnvInfo(BaseModel):
    interpreter_path: str | None
    interpreter_version: str | None
    name: str | None = "exponent"
    provider: Literal["venv", "pyenv", "pipenv", "conda"] | None = "pyenv"


class PortInfo(BaseModel):
    process_name: str
    port: int
    protocol: str
    pid: int | None
    uptime_seconds: float | None


class SystemInfo(BaseModel):
    name: str
    cwd: str
    os: str
    shell: str
    git: GitInfo | None
    python_env: PythonEnvInfo | None
    port_usage: list[PortInfo] | None = None


class HeartbeatInfo(BaseModel):
    exponent_version: str | None = None
    editable_installation: bool = False
    system_info: SystemInfo | None
    timestamp: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(datetime.UTC))
    timestamp_received: datetime.datetime | None = None
    cli_uuid: str | None = None
    config_dir: str | None = None


class RemoteFile(BaseModel):
    file_path: str
    working_directory: str = "."

    @cached_property
    def pure_path(self) -> PurePath:
        return PurePath(self.working_directory, self.file_path)

    @cached_property
    def path(self) -> Path:
        return Path(self.working_directory, self.file_path)

    @cached_property
    def name(self) -> str:
        return self.pure_path.name

    @cached_property
    def absolute_path(self) -> str:
        return self.path.absolute().as_posix()

    async def resolve(self, client_working_directory: str) -> str:
        working_directory = AsyncPath(self.working_directory, self.file_path)

        if not working_directory.is_absolute():
            working_directory = AsyncPath(client_working_directory, working_directory)

        return str(await working_directory.resolve())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, RemoteFile):
            return False

        return self.path.name == other.path.name

    def __lt__(self, other: "RemoteFile") -> bool:
        # Prefer shorter paths
        if (cmp := self._cmp_path_len(other)) is not None:
            return cmp

        # Prefer paths sorted by parent directory
        if (cmp := self._cmp_path_str(other)) is not None:
            return cmp

        # Prefer paths with alphabetical first character
        return self._cmp_first_char(other)

    def __hash__(self) -> int:
        return hash(self.absolute_path)

    def _cmp_first_char(self, other: "RemoteFile") -> bool:
        return self._cmp_str(self.path.name, other.path.name)

    def _cmp_path_len(self, other: "RemoteFile") -> bool | None:
        self_parts = self.path.absolute().parent.parts
        other_parts = other.path.absolute().parent.parts

        if len(self_parts) == len(other_parts):
            return None

        return len(self_parts) < len(other_parts)

    def _cmp_path_str(self, other: "RemoteFile") -> bool | None:
        self_parts = self.path.absolute().parent.parts
        other_parts = other.path.absolute().parent.parts

        if self_parts == other_parts:
            return None

        for a, b in zip(self_parts, other_parts):
            if a != b:
                return self._cmp_str(a, b)

        return False

    @staticmethod
    def _cmp_str(s1: str, s2: str) -> bool:
        if s1[:1].isalpha() == s2[:1].isalpha():
            return s1 < s2

        return s1[:1].isalpha()


class URLAttachment(BaseModel):
    attachment_type: Literal["url"] = "url"
    url: str
    content: str


class FileAttachment(BaseModel):
    attachment_type: Literal["file"] = "file"
    file: RemoteFile
    content: str
    truncated: bool = False


class TableSchemaAttachment(BaseModel):
    attachment_type: Literal["table_schema"] = "table_schema"
    table_name: str
    table_schema: dict[str, Any]


class PromptAttachment(BaseModel):
    attachment_type: Literal["prompt"] = "prompt"
    prompt_name: str
    prompt_content: str


class SQLAttachment(BaseModel):
    attachment_type: Literal["sql"] = "sql"
    query_content: str
    query_id: str


class SkillAttachment(BaseModel):
    attachment_type: Literal["skill"] = "skill"
    skill_name: str


MessageAttachment = Annotated[
    FileAttachment | URLAttachment | TableSchemaAttachment | PromptAttachment | SQLAttachment | SkillAttachment,
    Field(discriminator="attachment_type"),
]


class GitFileChange(BaseModel):
    path: str
    lines_added: int
    lines_deleted: int


class GitDiff(BaseModel):
    files: list[GitFileChange]
    truncated: bool = False  # True if there were more files than the limit
    total_files: int  # Total number of files changed, even if truncated


class GitCommitMetadata(BaseModel):
    author_name: str
    author_email: str
    author_date: str
    commit_date: str
    commit_message: str
    branch: str


class ChatMode(str, Enum):
    DEFAULT = "DEFAULT"  # chat just with model
    CLI = "CLI"
    CLOUD = "CLOUD"  # chat with cloud devbox
    CLOUD_SETUP = "CLOUD_SETUP"  # cloud environment setup
    CLOUD_SETUP_AUTO = "CLOUD_SETUP_AUTO"  # autonomous cloud environment setup
    CODEBASE = "CODEBASE"  # chat with codebase
    DATABASE = "DATABASE"  # chat with database connection
    WORKFLOW = "WORKFLOW"
    PLAYGROUND = "PLAYGROUND"  # playground mode with MCP tools only
    ONCALL = "ONCALL"  # incident response mode with auto-configured Datadog/Sentry/Sandbox

    @classmethod
    def requires_cli(cls, mode: "ChatMode") -> bool:
        return mode not in [
            cls.DATABASE,
            cls.CLOUD_SETUP,
            cls.CLOUD_SETUP_AUTO,
            cls.PLAYGROUND,
        ]


class ChatSource(str, Enum):
    CLI_SHELL = "CLI_SHELL"
    CLI_RUN = "CLI_RUN"
    WEB = "WEB"
    DESKTOP_APP = "DESKTOP_APP"
    VSCODE_EXTENSION = "VSCODE_EXTENSION"
    SLACK_APP = "SLACK_APP"
    SENTRY_APP = "SENTRY_APP"
    GITHUB_APP = "GITHUB_APP"


class AgentSubtype(str, Enum):
    CODING_AGENT = "CODING_AGENT"
    SRE_AGENT = "SRE_AGENT"


class CLIConnectedState(BaseModel):
    chat_uuid: str
    connected: bool
    last_connected_at: datetime.datetime | None
    connection_latency_ms: int | None
    system_info: SystemInfo | None
    exponent_version: str | None = None
    editable_installation: bool = False
    config_dir: str | None = None


class DevboxConnectedState(str, Enum):
    # The chat has been initialized, but the devbox is still loading
    DEVBOX_LOADING = "DEVBOX_LOADING"
    # CLI is connected and running on devbox
    CONNECTED = "CONNECTED"
    # Devbox has an error
    DEVBOX_ERROR = "DEVBOX_ERROR"
    # Devbox is going to idle
    PAUSING = "PAUSING"
    # Devbox has been paused and is not running
    PAUSED = "PAUSED"
    # Dev box is starting up. Sandbox exists but devbox is not running
    RESUMING = "RESUMING"


class CloudConnectedState(BaseModel):
    chat_uuid: str
    connected_state: DevboxConnectedState
    last_connected_at: datetime.datetime | None
    system_info: SystemInfo | None


class CLIErrorLog(BaseModel):
    event_data: str
    timestamp: datetime.datetime = datetime.datetime.now()
    attachment_data: str | None = None
    version: str | None = None
    chat_uuid: str | None = None

    @property
    def loaded_event_data(self) -> Any | None:
        try:
            return json.loads(self.event_data)
        except json.JSONDecodeError:
            return None

    @property
    def attachment_bytes(self) -> bytes | None:
        if not self.attachment_data:
            return None
        return self.attachment_data.encode()



import shutil
import subprocess
import sys
import time
from dataclasses import dataclass
from datetime import datetime
from typing import Any, cast

import click
import questionary

from exponent.commands.common import redirect_to_login, run_until_complete
from exponent.commands.settings import use_settings
from exponent.commands.types import exponent_cli_group
from exponent.core.config import Settings
from exponent.core.graphql.client import GraphQLClient
from exponent.core.graphql.generated_client.chats import (
    ChatsChatsChats,
    ChatsChatsChatsChats,
    ChatsChatsChatsChatsPrStatusPRInfo,
    ChatsChatsUnauthenticatedError,
)
from exponent.core.graphql.generated_client.enums import PRStatus


@exponent_cli_group()
def checkout_cli() -> None:
    """Checkout commands for syncing cloud changes to local."""
    pass


async def fetch_chats(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
) -> list[ChatsChatsChatsChats]:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    result = await graphql_client.get_chats()

    if isinstance(result.chats, ChatsChatsUnauthenticatedError):
        click.secho(f"Error: {result.chats.message}", fg="red")
        sys.exit(1)

    if isinstance(result.chats, ChatsChatsChats):
        return result.chats.chats

    click.secho("Unexpected response from server", fg="red")
    sys.exit(1)


def sort_chats_by_recency(
    chats: list[ChatsChatsChatsChats],
) -> list[ChatsChatsChatsChats]:
    def parse_updated_at(chat: ChatsChatsChatsChats) -> datetime:
        updated_at = chat.updated_at
        if updated_at is None:
            return datetime.min
        if isinstance(updated_at, datetime):
            return updated_at
        try:
            return datetime.fromisoformat(str(updated_at).replace("Z", "+00:00"))
        except (ValueError, TypeError):
            return datetime.min

    return sorted(chats, key=parse_updated_at, reverse=True)


@dataclass
class PRChoice:
    chat: ChatsChatsChatsChats
    pr_number: int
    pr_title: str


def get_all_open_prs(chat: ChatsChatsChatsChats) -> list[tuple[int, str]]:
    return [
        (pr.number, pr.title)
        for pr in chat.pr_status
        if isinstance(pr, ChatsChatsChatsChatsPrStatusPRInfo) and pr.status == PRStatus.OPEN
    ]


def collect_all_pr_choices(chats: list[ChatsChatsChatsChats]) -> list[PRChoice]:
    choices: list[PRChoice] = []
    for chat in chats:
        for pr_number, pr_title in get_all_open_prs(chat):
            choices.append(PRChoice(chat=chat, pr_number=pr_number, pr_title=pr_title))
    return choices


def format_pr_choice(pr_choice: PRChoice) -> dict[str, Any]:
    display = f"#{pr_choice.pr_number} {pr_choice.pr_title}"
    return {"name": display, "value": pr_choice}


def check_gh_installed() -> None:
    if shutil.which("gh") is None:
        click.secho("Error: GitHub CLI (gh) is not installed.", fg="red")
        click.echo("Please install it from: https://cli.github.com/")
        sys.exit(1)


MAX_PRS_TO_SHOW = 20
DEFAULT_POLL_INTERVAL = 3


def select_pr_interactive(pr_choices: list[PRChoice]) -> PRChoice:
    if not pr_choices:
        click.secho("No open PRs found.", fg="yellow")
        sys.exit(1)

    if len(pr_choices) == 1:
        selected = pr_choices[0]
        click.secho(f"Using PR: #{selected.pr_number} {selected.pr_title}", fg="cyan")
        return selected

    prs_to_show = pr_choices[:MAX_PRS_TO_SHOW]
    choices = [format_pr_choice(pr) for pr in prs_to_show]

    selected = questionary.select(
        "Select a PR to checkout:",
        choices=choices,
        style=questionary.Style(
            [
                ("question", "bold"),
                ("pointer", "fg:#33ccff bold"),
                ("highlighted", "fg:#33ccff"),
            ]
        ),
    ).ask()

    if selected is None:
        click.secho("\nCancelled", fg="yellow")
        sys.exit(0)

    return cast(PRChoice, selected)


def find_prs_by_chat_id(pr_choices: list[PRChoice], chat_id: str) -> list[PRChoice]:
    return [pr for pr in pr_choices if chat_id == str(pr.chat.chat_uuid)]


def get_local_head() -> str | None:
    result = subprocess.run(
        ["git", "rev-parse", "HEAD"],
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode == 0:
        return result.stdout.strip()
    return None


def get_current_branch() -> str | None:
    result = subprocess.run(
        ["git", "rev-parse", "--abbrev-ref", "HEAD"],
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode == 0:
        return result.stdout.strip()
    return None


def get_pr_branch(pr_number: int) -> str | None:
    result = subprocess.run(
        [
            "gh",
            "pr",
            "view",
            str(pr_number),
            "--json",
            "headRefName",
            "--jq",
            ".headRefName",
        ],
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode == 0:
        return result.stdout.strip()
    return None


def get_remote_head(pr_number: int) -> str | None:
    result = subprocess.run(
        [
            "gh",
            "pr",
            "view",
            str(pr_number),
            "--json",
            "headRefOid",
            "--jq",
            ".headRefOid",
        ],
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode == 0:
        return result.stdout.strip()
    return None


def get_uncommitted_files() -> list[str]:
    result = subprocess.run(
        ["git", "status", "--porcelain"],
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode == 0 and result.stdout.strip():
        return [line[3:] for line in result.stdout.strip().split("\n") if line]
    return []


def get_local_commits_on_pr_branch(pr_number: int) -> list[str]:
    current_branch = get_current_branch()
    pr_branch = get_pr_branch(pr_number)
    if current_branch != pr_branch:
        return []
    remote_head = get_remote_head(pr_number)
    if not remote_head:
        return []
    result = subprocess.run(
        ["git", "log", f"{remote_head}..HEAD", "--oneline"],
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode == 0 and result.stdout.strip():
        return result.stdout.strip().split("\n")
    return []


def force_checkout_pr(pr_number: int) -> bool:
    result = subprocess.run(
        ["gh", "pr", "checkout", str(pr_number), "--force"],
        check=False,
    )
    return result.returncode == 0


def show_local_changes_and_confirm(pr_number: int) -> bool:
    uncommitted = get_uncommitted_files()
    local_commits = get_local_commits_on_pr_branch(pr_number)

    if not uncommitted and not local_commits:
        return True

    click.secho("\nLocal changes would be lost:", fg="yellow")

    if uncommitted:
        click.echo("\nUncommitted files:")
        for f in uncommitted[:10]:
            click.echo(f"  {f}")
        if len(uncommitted) > 10:
            click.echo(f"  ... and {len(uncommitted) - 10} more")

    if local_commits:
        click.echo("\nLocal commits:")
        for c in local_commits[:10]:
            click.echo(f"  {c}")
        if len(local_commits) > 10:
            click.echo(f"  ... and {len(local_commits) - 10} more")

    return click.confirm("\nForce checkout and discard these changes?")


def live_sync_pr(pr_number: int, poll_interval: int) -> None:
    last_local_head = get_local_head()
    click.echo(f"\nLive syncing PR #{pr_number} (polling every {poll_interval}s, Ctrl+C to stop)...")
    click.echo("Will exit if local changes are detected.\n")

    try:
        while True:
            time.sleep(poll_interval)

            uncommitted = get_uncommitted_files()
            if uncommitted:
                click.secho(
                    f"\nLocal changes detected ({len(uncommitted)} files), stopping live sync",
                    fg="yellow",
                )
                break

            current_local_head = get_local_head()
            if current_local_head != last_local_head:
                click.secho("\nLocal commits detected, stopping live sync", fg="yellow")
                break

            remote_head = get_remote_head(pr_number)
            if remote_head is None:
                click.secho("Could not fetch remote head, PR may be closed", fg="yellow")
                break

            if current_local_head != remote_head:
                click.echo(f"New commits detected ({remote_head[:7]}), syncing...")
                if force_checkout_pr(pr_number):
                    last_local_head = get_local_head()
                    click.secho("Synced successfully", fg="green")
                else:
                    click.secho("Sync failed", fg="red")
                    break
    except KeyboardInterrupt:
        click.echo("\nStopped live sync")


@checkout_cli.command(hidden=True)
@click.argument("chat_id", required=False)
@click.option(
    "-l",
    "--live",
    is_flag=True,
    help="Live mode: keep syncing with remote until local changes are detected",
)
@click.option(
    "--poll-interval",
    default=DEFAULT_POLL_INTERVAL,
    help="Seconds between polls in live mode",
)
@use_settings
def checkout(
    settings: Settings,
    chat_id: str | None = None,
    live: bool = False,
    poll_interval: int = DEFAULT_POLL_INTERVAL,
) -> None:
    """Checkout a PR from a cloud chat locally.

    If CHAT_ID is provided, shows PRs for that specific chat.
    Otherwise, presents an interactive selector of all open PRs.
    """
    if not settings.api_key:
        redirect_to_login(settings)
        return

    api_key = settings.api_key
    base_api_url = settings.get_base_api_url()
    base_ws_url = settings.get_base_ws_url()

    chats = run_until_complete(fetch_chats(api_key, base_api_url, base_ws_url))

    sorted_chats = sort_chats_by_recency(chats)
    all_pr_choices = collect_all_pr_choices(sorted_chats)

    if not all_pr_choices:
        click.secho("No chats with open PRs found.", fg="yellow")
        sys.exit(1)

    if chat_id:
        matching_prs = find_prs_by_chat_id(all_pr_choices, chat_id)
        if not matching_prs:
            click.secho(f"No open PRs found for chat: {chat_id}", fg="red")
            sys.exit(1)
        selected_pr = select_pr_interactive(matching_prs)
    else:
        selected_pr = select_pr_interactive(all_pr_choices)

    check_gh_installed()
    pr_number = selected_pr.pr_number
    pr_title = selected_pr.pr_title
    chat_url = f"{settings.base_url}/chats/{selected_pr.chat.chat_uuid}"
    click.secho(f"\nSelected: #{pr_number} {pr_title}", fg="green")
    click.echo(f"Chat: {chat_url}")

    if not show_local_changes_and_confirm(pr_number):
        click.echo("Cancelled")
        sys.exit(0)

    click.echo(f"\nChecking out PR #{pr_number}...")
    if not force_checkout_pr(pr_number):
        click.secho("Checkout failed", fg="red")
        sys.exit(1)

    click.secho("Checkout successful", fg="green")

    if live:
        live_sync_pr(pr_number, poll_interval)



import base64
import importlib
from typing import TYPE_CHECKING, Any

import msgspec

from exponent.core.remote_execution.file_reference import FilePath

if TYPE_CHECKING:
    from exponent_server.core.tools.edit_tool import (
        EditToolInput as ServerSideEditToolInput,
    )


def _load_format_as_xml() -> Any:
    try:
        return importlib.import_module("pydantic_ai.format_prompt").format_as_xml
    except (ModuleNotFoundError, AttributeError):
        return getattr(importlib.import_module("pydantic_ai.format_as_xml"), "format_as_xml")


format_as_xml = _load_format_as_xml()


class PartialToolResult(msgspec.Struct, tag_field="tool_name", omit_defaults=True):
    pass


class ToolInput(msgspec.Struct, tag_field="tool_name", omit_defaults=True):
    """Concrete subclasses describe the full input schema for a tool."""

    def to_llm(self) -> dict[str, Any]:
        """Convert ToolInput to LLM-friendly typed dict format.

        Returns a dictionary with the tool parameters, excluding the tool_name
        which is handled separately by the LLM integration layer.

        Returns:
            self by default, which msgspec will serialize appropriately.
        """
        return msgspec.to_builtins(self)

    def parallelism_key(self) -> str | None:
        """Return a key for parallelism control.

        Tools with the same non-None parallelism key will be executed serially,
        while tools with different keys (or None) can run in parallel.

        By default, returns None which means no parallelism restrictions.
        Subclasses can override this to implement custom parallelism behavior.

        Returns:
            A string key for grouping serial execution, or None for unrestricted parallelism.
        """
        return None


class ToolResult(msgspec.Struct, tag_field="tool_name", omit_defaults=True):
    """Concrete subclasses return data from a tool execution."""

    def to_text(self) -> str:
        """
        This provides a default textual representation of the tool result. Override it as needed for your tool."""
        d = msgspec.to_builtins(self)
        del d["tool_name"]
        return format_as_xml(d, include_root_tag=False, item_tag="item")


class ErrorToolResult(ToolResult, tag="error"):
    error_message: str
    is_assistant_error: bool = False


READ_TOOL_NAME = "read"
READ_TOOL_ARTIFACT_NAME = "read_tool_artifact"


class ReadToolInput(ToolInput, tag=READ_TOOL_NAME):
    file_path: FilePath
    offset: int | None = None
    limit: int | None = None


class FileMetadata(msgspec.Struct):
    modified_timestamp: float
    file_mode: str


class ReadToolArtifactResult(ToolResult, tag=READ_TOOL_ARTIFACT_NAME):
    s3_uri: str
    file_path: FilePath
    media_type: str

    def to_text(self) -> str:
        return f"[Image artifact uploaded to {self.s3_uri}]"

    async def get_base64_content(self, chat_uuid: str) -> str:
        """Download the artifact from S3 and return as base64-encoded string.

        Args:
            chat_uuid: The chat UUID for S3 client scoping

        Returns:
            Base64-encoded string of the file content
        """
        from exponent_server.core.services.s3_services.s3 import S3Client

        s3_client = S3Client(chat_uuid=chat_uuid)
        await s3_client.authenticate()

        s3_uri = self.s3_uri
        bucket_prefix = f"s3://{s3_client.bucket_name}/"
        if s3_uri.startswith(bucket_prefix):
            s3_key = s3_uri[len(bucket_prefix) :]
            chat_prefix = f"{chat_uuid}/"
            if s3_key.startswith(chat_prefix):
                s3_key = s3_key[len(chat_prefix) :]
        else:
            raise ValueError(f"S3 URI {s3_uri} does not match expected bucket prefix {bucket_prefix}")

        file_bytes = await s3_client.get_object(s3_key)
        return base64.standard_b64encode(file_bytes).decode("utf-8")


class ReadToolResult(ToolResult, tag=READ_TOOL_NAME):
    content: str | None = None
    num_lines: int | None = None
    start_line: int | None = None
    total_lines: int | None = None
    metadata: FileMetadata | None = None
    artifact: ReadToolArtifactResult | None = None

    def to_text(self) -> str:
        if self.artifact:
            return self.artifact.to_text()
        assert self.content is not None
        assert self.start_line is not None

        metadata_attrs = []
        if self.total_lines is not None:
            metadata_attrs.append(f'total_lines="{self.total_lines}"')
        if self.metadata is not None:
            metadata_attrs.append(f'mtime="{self.metadata.modified_timestamp}"')
            metadata_attrs.append(f'mode="{self.metadata.file_mode}"')

        lines = self.content.splitlines()
        lines = [f"{str(i).rjust(6)}→{line}" for i, line in enumerate(lines, start=self.start_line + 1)]

        if metadata_attrs:
            metadata_header = "<metadata " + " ".join(metadata_attrs) + " />"
            return metadata_header + "\n" + "\n".join(lines)
        return "\n".join(lines)


GLOB_TOOL_NAME = "glob"


class GlobToolInput(ToolInput, tag=GLOB_TOOL_NAME):
    pattern: str
    path: str | None = None


class GlobToolResult(ToolResult, tag=GLOB_TOOL_NAME):
    filenames: list[str]
    duration_ms: int
    num_files: int
    truncated: bool


WRITE_TOOL_NAME = "write"


class WriteToolInput(ToolInput, tag=WRITE_TOOL_NAME):
    file_path: FilePath
    content: str


class WriteToolResult(ToolResult, tag=WRITE_TOOL_NAME):
    message: str
    metadata: FileMetadata | None = None


GREP_TOOL_NAME = "grep"


class GrepToolInput(ToolInput, tag=GREP_TOOL_NAME):
    pattern: str
    path: str | None = None
    include: str | None = None
    multiline: bool | None = None


class GrepToolResult(ToolResult, tag=GREP_TOOL_NAME):
    matches: list[str]
    truncated: bool = False


EDIT_TOOL_NAME = "edit"


# This is only used in the CLI. The server side type is edit_tool.py
class EditToolInput(ToolInput, tag=EDIT_TOOL_NAME):
    file_path: FilePath
    old_string: str
    new_string: str
    replace_all: bool = False
    # When set, matches everything from old_string up to and including old_string_end.
    # Cannot be used with replace_all=True.
    old_string_end: str | None = None
    # The last retrieved modified timestamp. This is used to check if the file has been modified since the last read/write that we are aware of (in which case we should re-read the file before editing).
    last_known_modified_timestamp: float | None = None

    @classmethod
    def from_server_side_input(
        cls,
        server_side_input: "ServerSideEditToolInput",
        last_known_modified_timestamp: float | None,
    ) -> "EditToolInput":
        return cls(
            file_path=server_side_input.file_path,
            old_string=server_side_input.old_string,
            new_string=server_side_input.new_string,
            replace_all=server_side_input.replace_all,
            old_string_end=server_side_input.old_string_end,
            last_known_modified_timestamp=last_known_modified_timestamp,
        )


class EditToolResult(ToolResult, tag=EDIT_TOOL_NAME):
    message: str
    metadata: FileMetadata | None = None


BASH_TOOL_NAME = "bash"


class BashToolInput(ToolInput, tag=BASH_TOOL_NAME):
    command: str
    timeout: int | None = None
    description: str | None = None
    background: bool = False


class PartialBashToolResult(PartialToolResult, tag=BASH_TOOL_NAME):
    shell_output: str | None = None


class BashToolResult(ToolResult, tag=BASH_TOOL_NAME):
    shell_output: str
    duration_ms: int
    exit_code: int | None
    timed_out: bool
    stopped_by_user: bool
    pid: int | None = None
    output_file: str | None = None


DOWNLOAD_ARTIFACT_TOOL_NAME = "download_artifact"


class DownloadArtifactToolInput(ToolInput, tag=DOWNLOAD_ARTIFACT_TOOL_NAME):
    presigned_url: str
    file_path: FilePath
    artifact_id: str
    overwrite: bool = True


class DownloadArtifactToolResult(ToolResult, tag=DOWNLOAD_ARTIFACT_TOOL_NAME):
    file_path: FilePath
    artifact_id: str
    file_size_bytes: int
    content_preview: str | None = None
    num_lines: int | None = None
    total_lines: int | None = None
    truncated: bool = False


UPLOAD_ARTIFACT_TOOL_NAME = "upload_artifact"


class UploadArtifactToolInput(ToolInput, tag=UPLOAD_ARTIFACT_TOOL_NAME):
    presigned_url: str
    file_path: FilePath
    artifact_id: str
    content_type: str


class UploadArtifactToolResult(ToolResult, tag=UPLOAD_ARTIFACT_TOOL_NAME):
    artifact_id: str
    file_size_bytes: int
    content_type: str


STORE_ARTIFACT_TOOL_NAME = "store_artifact"


class StoreArtifactToolInput(ToolInput, tag=STORE_ARTIFACT_TOOL_NAME):
    """Store an artifact to the CLI's config directory.

    The CLI resolves the local path using its own INDENT_HOME configuration,
    storing files flat in ~/.indent/chats/{chat_uuid}/{filename}.

    The chat_uuid is obtained from the CLI's session context, not passed in the input.
    """

    presigned_url: str
    filename: str  # Just the filename, no path (e.g., "document.pdf")


class StoreArtifactToolResult(ToolResult, tag=STORE_ARTIFACT_TOOL_NAME):
    local_path: str  # Full path where CLI stored the file
    file_size_bytes: int


class HttpRequest(msgspec.Struct, tag="http_fetch_cli"):
    url: str
    method: str = "GET"
    headers: dict[str, str] | None = None
    timeout: int | None = None


class HttpResponse(msgspec.Struct, tag="http_fetch_cli"):
    status_code: int | None = None
    content: str | None = None
    error_message: str | None = None
    duration_ms: int | None = None
    headers: dict[str, str] | None = None


ToolInputType = (
    ReadToolInput
    | WriteToolInput
    | GlobToolInput
    | GrepToolInput
    | EditToolInput
    | BashToolInput
    | DownloadArtifactToolInput
    | UploadArtifactToolInput
    | StoreArtifactToolInput
)
PartialToolResultType = PartialBashToolResult

ToolResultType = (
    ReadToolResult
    | ReadToolArtifactResult
    | WriteToolResult
    | GlobToolResult
    | GrepToolResult
    | EditToolResult
    | BashToolResult
    | DownloadArtifactToolResult
    | UploadArtifactToolResult
    | StoreArtifactToolResult
    | ErrorToolResult
)


class ToolExecutionRequest(msgspec.Struct, tag="tool_execution"):
    tool_input: ToolInputType


class GetAllFilesRequest(msgspec.Struct, tag="get_all_files"):
    pass


class TerminateRequest(msgspec.Struct, tag="terminate"):
    pass


class SwitchCLIChatRequest(msgspec.Struct, tag="switch_cli_chat"):
    new_chat_uuid: str


# This message is sent periodically from the client to keep the connection alive while the chat is turning.
# This synergizes with CLI-side timeouts to avoid disconnecting during long operations.
class KeepAliveCliChatRequest(msgspec.Struct, tag="keep_alive_cli_chat"):
    pass


class BatchToolExecutionRequest(msgspec.Struct, tag="batch_tool_execution"):
    tool_inputs: list[ToolInputType]


class GetAllFilesResponse(msgspec.Struct, tag="get_all_files"):
    files: list[str]


class TerminateResponse(msgspec.Struct, tag="terminate"):
    pass


class TimeoutResponse(msgspec.Struct, tag="timeout"):
    pass


class BatchToolExecutionResponse(msgspec.Struct, tag="batch_tool_execution"):
    tool_results: list[ToolResultType]


class SwitchCLIChatResponse(msgspec.Struct, tag="switch_cli_chat"):
    pass


class KeepAliveCliChatResponse(msgspec.Struct, tag="keep_alive_cli_chat"):
    pass


class GenerateUploadUrlRequest(msgspec.Struct, tag="generate_upload_url"):
    s3_key: str
    content_type: str


class GenerateUploadUrlResponse(msgspec.Struct, tag="generate_upload_url"):
    upload_url: str
    s3_uri: str


class DownloadFromUrlRequest(msgspec.Struct, tag="download_from_url"):
    """Request to download a file from a URL to the local filesystem."""

    url: str
    file_path: str
    timeout: int = 300


class DownloadFromUrlResponse(msgspec.Struct, tag="download_from_url"):
    """Response after downloading a file from a URL."""

    file_path: str
    file_size_bytes: int
    success: bool
    error_message: str | None = None


# Terminal session management
class StartTerminalRequest(msgspec.Struct, tag="start_terminal"):
    """Start a new terminal session with PTY"""

    session_id: str
    command: list[str] | None = None  # None = default shell, or specific command
    cols: int = 80
    rows: int = 24
    env: dict[str, str] | None = None  # Additional environment variables
    name: str | None = None  # Display name for the terminal


class StartTerminalResponse(msgspec.Struct, tag="start_terminal"):
    """Response after starting terminal"""

    session_id: str
    success: bool
    error_message: str | None = None
    name: str | None = None  # Display name for the terminal


# Terminal input (user typing)
class TerminalInputRequest(msgspec.Struct, tag="terminal_input"):
    """Send user input to terminal"""

    session_id: str
    data: str  # Raw input data from user


class TerminalInputResponse(msgspec.Struct, tag="terminal_input"):
    """Acknowledge input received"""

    session_id: str
    success: bool
    error_message: str | None = None


# Terminal resize
class TerminalResizeRequest(msgspec.Struct, tag="terminal_resize"):
    """Resize terminal dimensions"""

    session_id: str
    cols: int
    rows: int


class TerminalResizeResponse(msgspec.Struct, tag="terminal_resize"):
    """Acknowledge resize"""

    session_id: str
    success: bool
    error_message: str | None = None


# Terminal stop
class StopTerminalRequest(msgspec.Struct, tag="stop_terminal"):
    """Stop a terminal session"""

    session_id: str


class StopTerminalResponse(msgspec.Struct, tag="stop_terminal"):
    """Acknowledge terminal stopped"""

    session_id: str
    success: bool
    error_message: str | None = None


class ListTerminalsRequest(msgspec.Struct, tag="list_terminals"):
    """Request to list all active terminal sessions"""

    pass


class TerminalInfo(msgspec.Struct):
    """Information about a terminal session"""

    session_id: str
    name: str | None = None
    running: bool = True


class ListTerminalsResponse(msgspec.Struct, tag="list_terminals"):
    """Response with list of active terminal sessions"""

    terminals: list[TerminalInfo]


class StreamingCodeExecutionRequest(msgspec.Struct, tag="streaming_code_execution"):
    correlation_id: str
    language: str  # "python" or "shell"
    content: str
    timeout: int


class CliRpcRequest(msgspec.Struct):
    request_id: str
    request: (
        ToolExecutionRequest
        | GetAllFilesRequest
        | TerminateRequest
        | HttpRequest
        | BatchToolExecutionRequest
        | SwitchCLIChatRequest
        | KeepAliveCliChatRequest
        | GenerateUploadUrlRequest
        | DownloadFromUrlRequest
        | StartTerminalRequest
        | TerminalInputRequest
        | TerminalResizeRequest
        | StopTerminalRequest
        | ListTerminalsRequest
        | StreamingCodeExecutionRequest
    )


class ToolExecutionResponse(msgspec.Struct, tag="tool_execution"):
    tool_result: ToolResultType


class ErrorResponse(msgspec.Struct, tag="error"):
    error_message: str


class StreamingCodeExecutionResponseChunk(msgspec.Struct, tag="streaming_code_execution_chunk"):
    correlation_id: str
    content: str
    truncated: bool = False

    def add(self, new_chunk: "StreamingCodeExecutionResponseChunk") -> "StreamingCodeExecutionResponseChunk":
        """Aggregates content of this and a new chunk."""
        assert self.correlation_id == new_chunk.correlation_id
        return StreamingCodeExecutionResponseChunk(
            correlation_id=self.correlation_id,
            content=self.content + new_chunk.content,
            truncated=self.truncated or new_chunk.truncated,
        )


class StreamingCodeExecutionResponse(msgspec.Struct, tag="streaming_code_execution"):
    correlation_id: str
    content: str
    truncated: bool = False
    # Only present for shell code execution
    cancelled_for_timeout: bool = False
    exit_code: int | None = None
    halted: bool = False
    output_file: str | None = None


class StreamingErrorResponse(msgspec.Struct, tag="streaming_error"):
    correlation_id: str
    error_message: str


# WebSocket Authentication Messages
class AuthMessage(msgspec.Struct):
    """Authentication message sent over websocket."""

    api_key: str | None = None
    token: str | None = None


class AuthSuccessMessage(msgspec.Struct):
    """Authentication success response."""

    pass


class AuthFailedMessage(msgspec.Struct):
    """Authentication failure response."""

    reason: str


class BackgroundProcessCompletedNotification(msgspec.Struct, tag="background_process_completed"):
    """Sent from CLI to server when a background bash process completes."""

    pid: int
    command: str
    exit_code: int
    output: str
    output_file: str
    correlation_id: str
    truncated: bool = False
    duration_ms: int | None = None


CliNotificationType = BackgroundProcessCompletedNotification


class CliRpcResponse(msgspec.Struct):
    request_id: str
    response: (
        ToolExecutionResponse
        | GetAllFilesResponse
        | ErrorResponse
        | TerminateResponse
        | TimeoutResponse
        | BatchToolExecutionResponse
        | HttpResponse
        | SwitchCLIChatResponse
        | KeepAliveCliChatResponse
        | GenerateUploadUrlResponse
        | DownloadFromUrlResponse
        | StartTerminalResponse
        | TerminalInputResponse
        | TerminalResizeResponse
        | StopTerminalResponse
        | ListTerminalsResponse
        | StreamingCodeExecutionResponseChunk
        | StreamingCodeExecutionResponse
        | StreamingErrorResponse
    )



from exponent.core.remote_execution.client.client import (
    RemoteExecutionClient as RemoteExecutionClient,
)
from exponent.core.remote_execution.client.types import (
    REMOTE_EXECUTION_CLIENT_EXIT_INFO as REMOTE_EXECUTION_CLIENT_EXIT_INFO,
    SwitchCLIChat as SwitchCLIChat,
    TModel as TModel,
    WSDisconnected as WSDisconnected,
    cli_uuid as cli_uuid,
)

__all__ = [
    "REMOTE_EXECUTION_CLIENT_EXIT_INFO",
    "RemoteExecutionClient",
    "SwitchCLIChat",
    "TModel",
    "WSDisconnected",
    "cli_uuid",
]



from __future__ import annotations

import uuid
from dataclasses import dataclass
from typing import TypeVar

from pydantic import BaseModel


@dataclass
class WSDisconnected:
    error_message: str | None = None


@dataclass
class SwitchCLIChat:
    new_chat_uuid: str


REMOTE_EXECUTION_CLIENT_EXIT_INFO = WSDisconnected | SwitchCLIChat

TModel = TypeVar("TModel", bound=BaseModel)

cli_uuid = uuid.uuid4()



# Generated by ariadne-codegen
# Source: exponent/core/graphql/operations

from typing import Any, Optional, Union
from uuid import UUID

from .base_client import AsyncBaseClient
from .base_model import UNSET, UnsetType
from .chats import Chats
from .create_cloud_chat_from_repository import CreateCloudChatFromRepository
from .enable_cloud_repository import EnableCloudRepository
from .github_repositories import GithubRepositories
from .halt_chat_stream import HaltChatStream
from .input_types import ChatConfig, ChatInput, ChatResourceConfigInput, RepositoryInput
from .rebuild_cloud_repository import RebuildCloudRepository
from .refresh_api_key import RefreshApiKey
from .report_sandbox_info import ReportSandboxInfo
from .set_login_complete import SetLoginComplete
from .start_chat_turn import StartChatTurn


def gql(q: str) -> str:
    return q


class IndentGraphQLClient(AsyncBaseClient):
    async def chats(self, **kwargs: Any) -> Chats:
        query = gql("""
            query Chats {
              chats {
                __typename
                ... on UnauthenticatedError {
                  message
                }
                ... on Chats {
                  chats {
                    chatUuid
                    name
                    subtitle
                    isShared
                    isStarted
                    updatedAt
                    prStatus {
                      __typename
                      ... on PRInfo {
                        number
                        title
                        status
                      }
                    }
                  }
                }
              }
            }
            """)
        variables: dict[str, object] = {}
        response = await self.execute(
            query=query, operation_name="Chats", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return Chats.model_validate(data)

    async def create_cloud_chat_from_repository(
        self, resource_config: ChatResourceConfigInput, **kwargs: Any
    ) -> CreateCloudChatFromRepository:
        query = gql("""
            mutation CreateCloudChatFromRepository($resourceConfig: ChatResourceConfigInput!) {
              createChatWithResourceConfig(input: {resourceConfig: $resourceConfig}) {
                __typename
                ... on Chat {
                  chatUuid
                }
                ... on UnauthenticatedError {
                  message
                }
                ... on NotFoundError {
                  resourceType
                  resourceId
                  message
                }
                ... on InputValidationError {
                  message
                }
              }
            }
            """)
        variables: dict[str, object] = {"resourceConfig": resource_config}
        response = await self.execute(
            query=query,
            operation_name="CreateCloudChatFromRepository",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return CreateCloudChatFromRepository.model_validate(data)

    async def enable_cloud_repository(
        self, repositories: list[RepositoryInput], **kwargs: Any
    ) -> EnableCloudRepository:
        query = gql("""
            mutation EnableCloudRepository($repositories: [RepositoryInput!]!) {
              enableCloudRepository(input: {repositories: $repositories}) {
                __typename
                ... on EnableCloudRepositoriesResult {
                  results {
                    orgName
                    repoName
                    success
                    errorMessage
                    images {
                      buildRef
                      createdAt
                      updatedAt
                    }
                  }
                }
                ... on UnauthenticatedError {
                  message
                }
                ... on NotFoundError {
                  resourceType
                  resourceId
                  message
                }
                ... on CloudSessionError {
                  message
                }
              }
            }
            """)
        variables: dict[str, object] = {"repositories": repositories}
        response = await self.execute(
            query=query,
            operation_name="EnableCloudRepository",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return EnableCloudRepository.model_validate(data)

    async def github_repositories(self, **kwargs: Any) -> GithubRepositories:
        query = gql("""
            query GithubRepositories {
              githubRepositories {
                __typename
                ... on Repositories {
                  repositories {
                    uuid: repositoryUuid
                    githubOrgName
                    githubRepoName
                    createdAt
                    updatedAt
                  }
                }
              }
            }
            """)
        variables: dict[str, object] = {}
        response = await self.execute(
            query=query,
            operation_name="GithubRepositories",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return GithubRepositories.model_validate(data)

    async def halt_chat_stream(self, chat_uuid: UUID, **kwargs: Any) -> HaltChatStream:
        query = gql("""
            mutation HaltChatStream($chatUuid: UUID!) {
              haltChatStream(input: {chatUuid: $chatUuid}) {
                __typename
                ... on Chat {
                  chatUuid
                }
                ... on UnauthenticatedError {
                  message
                }
                ... on NotFoundError {
                  resourceType
                  resourceId
                  message
                }
                ... on RemoteExecutionError {
                  message
                }
              }
            }
            """)
        variables: dict[str, object] = {"chatUuid": chat_uuid}
        response = await self.execute(
            query=query, operation_name="HaltChatStream", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return HaltChatStream.model_validate(data)

    async def rebuild_cloud_repository(
        self, repository_uuid: UUID, **kwargs: Any
    ) -> RebuildCloudRepository:
        query = gql("""
            mutation RebuildCloudRepository($repositoryUuid: UUID!) {
              rebuildCloudRepository(input: {repositoryUuid: $repositoryUuid}) {
                __typename
                ... on ContainerImages {
                  images {
                    buildRef
                    createdAt
                    updatedAt
                  }
                }
                ... on UnauthenticatedError {
                  message
                }
                ... on NotFoundError {
                  resourceType
                  resourceId
                  message
                }
                ... on CloudSessionError {
                  message
                }
              }
            }
            """)
        variables: dict[str, object] = {"repositoryUuid": repository_uuid}
        response = await self.execute(
            query=query,
            operation_name="RebuildCloudRepository",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return RebuildCloudRepository.model_validate(data)

    async def refresh_api_key(self, **kwargs: Any) -> RefreshApiKey:
        query = gql("""
            mutation RefreshApiKey {
              refreshApiKey(input: {}) {
                __typename
                ... on User {
                  userApiKey
                }
                ... on UnauthenticatedError {
                  message
                }
              }
            }
            """)
        variables: dict[str, object] = {}
        response = await self.execute(
            query=query, operation_name="RefreshApiKey", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return RefreshApiKey.model_validate(data)

    async def report_sandbox_info(
        self,
        sandbox_id: str,
        disk_usage_gb: Union[Optional[float], UnsetType] = UNSET,
        indent_log_file: Union[Optional[str], UnsetType] = UNSET,
        **kwargs: Any
    ) -> ReportSandboxInfo:
        query = gql("""
            mutation ReportSandboxInfo($sandboxId: String!, $diskUsageGb: Float, $indentLogFile: String) {
              reportSandboxInfo(
                input: {sandboxId: $sandboxId, diskUsageGb: $diskUsageGb, indentLogFile: $indentLogFile}
              ) {
                __typename
                ... on SandboxInfoResponse {
                  success
                  message
                }
                ... on UnauthenticatedError {
                  message
                }
              }
            }
            """)
        variables: dict[str, object] = {
            "sandboxId": sandbox_id,
            "diskUsageGb": disk_usage_gb,
            "indentLogFile": indent_log_file,
        }
        response = await self.execute(
            query=query,
            operation_name="ReportSandboxInfo",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return ReportSandboxInfo.model_validate(data)

    async def set_login_complete(self, **kwargs: Any) -> SetLoginComplete:
        query = gql("""
            mutation SetLoginComplete {
              setLoginComplete(input: {}) {
                __typename
                ... on User {
                  userApiKey
                }
                ... on UnauthenticatedError {
                  message
                }
              }
            }
            """)
        variables: dict[str, object] = {}
        response = await self.execute(
            query=query,
            operation_name="SetLoginComplete",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return SetLoginComplete.model_validate(data)

    async def start_chat_turn(
        self,
        chat_input: ChatInput,
        chat_config: ChatConfig,
        parent_uuid: Union[Optional[UUID], UnsetType] = UNSET,
        **kwargs: Any
    ) -> StartChatTurn:
        query = gql("""
            mutation StartChatTurn($chatInput: ChatInput!, $parentUuid: UUID, $chatConfig: ChatConfig!) {
              startChatTurn(
                input: {chatInput: $chatInput, parentUuid: $parentUuid, chatConfig: $chatConfig}
              ) {
                __typename
                ... on UnauthenticatedError {
                  message
                }
                ... on NotFoundError {
                  resourceType
                  resourceId
                  message
                }
                ... on Chat {
                  chatUuid
                }
              }
            }
            """)
        variables: dict[str, object] = {
            "chatInput": chat_input,
            "parentUuid": parent_uuid,
            "chatConfig": chat_config,
        }
        response = await self.execute(
            query=query, operation_name="StartChatTurn", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return StartChatTurn.model_validate(data)



# Generated by ariadne-codegen
# Source: exponent/core/graphql/operations

from typing import Literal, Optional, Union
from uuid import UUID

from pydantic import Field

from .base_model import BaseModel


class CreateCloudChatFromRepository(BaseModel):
    create_chat_with_resource_config: Union[
        "CreateCloudChatFromRepositoryCreateChatWithResourceConfigChat",
        "CreateCloudChatFromRepositoryCreateChatWithResourceConfigNotFoundError",
        "CreateCloudChatFromRepositoryCreateChatWithResourceConfigInputValidationError",
        "CreateCloudChatFromRepositoryCreateChatWithResourceConfigPlanError",
        "CreateCloudChatFromRepositoryCreateChatWithResourceConfigUnauthenticatedError",
    ] = Field(alias="createChatWithResourceConfig", discriminator="typename__")


class CreateCloudChatFromRepositoryCreateChatWithResourceConfigChat(BaseModel):
    typename__: Literal["Chat"] = Field(alias="__typename")
    chat_uuid: UUID = Field(alias="chatUuid")


class CreateCloudChatFromRepositoryCreateChatWithResourceConfigNotFoundError(BaseModel):
    typename__: Literal["NotFoundError"] = Field(alias="__typename")
    resource_type: str = Field(alias="resourceType")
    resource_id: Optional[str] = Field(alias="resourceId")
    message: str


class CreateCloudChatFromRepositoryCreateChatWithResourceConfigInputValidationError(
    BaseModel
):
    typename__: Literal["InputValidationError"] = Field(alias="__typename")
    message: str


class CreateCloudChatFromRepositoryCreateChatWithResourceConfigPlanError(BaseModel):
    typename__: Literal["PlanError"] = Field(alias="__typename")


class CreateCloudChatFromRepositoryCreateChatWithResourceConfigUnauthenticatedError(
    BaseModel
):
    typename__: Literal["UnauthenticatedError"] = Field(alias="__typename")
    message: str


CreateCloudChatFromRepository.model_rebuild()



"""File download implementation for remote execution client."""

import logging
import os
from pathlib import Path

import httpx

from exponent.core.remote_execution.cli_rpc_types import (
    DownloadFromUrlRequest,
    DownloadFromUrlResponse,
)

logger = logging.getLogger(__name__)

DEFAULT_USER_AGENT = "Indent-HTTP-Client/1.0"


async def download_file_from_url(
    request: DownloadFromUrlRequest,
) -> DownloadFromUrlResponse:
    """
    Download a file from a URL to the local filesystem.

    Args:
        request: DownloadFromUrlRequest containing URL, file path, and timeout

    Returns:
        DownloadFromUrlResponse with file path, size, and success status
    """
    logger.info(f"Downloading file to {request.file_path}")

    try:
        file_path = Path(request.file_path)
        file_path.parent.mkdir(parents=True, exist_ok=True)

        headers = {"User-Agent": DEFAULT_USER_AGENT}

        async with httpx.AsyncClient(timeout=request.timeout) as client:
            async with client.stream("GET", request.url, headers=headers) as response:
                response.raise_for_status()
                with open(file_path, "wb") as f:
                    async for chunk in response.aiter_bytes():
                        f.write(chunk)

            file_size = os.path.getsize(file_path)

            logger.info(f"Downloaded file to {request.file_path} ({file_size} bytes)")

            return DownloadFromUrlResponse(
                file_path=request.file_path,
                file_size_bytes=file_size,
                success=True,
                error_message=None,
            )

    except httpx.TimeoutException:
        error_msg = f"Download from {request.url} timed out after {request.timeout} seconds"
        logger.error(error_msg)
        return DownloadFromUrlResponse(
            file_path=request.file_path,
            file_size_bytes=0,
            success=False,
            error_message=error_msg,
        )

    except httpx.HTTPStatusError as e:
        error_msg = f"HTTP error downloading from {request.url}: {e.response.status_code}"
        logger.error(error_msg)
        return DownloadFromUrlResponse(
            file_path=request.file_path,
            file_size_bytes=0,
            success=False,
            error_message=error_msg,
        )

    except httpx.RequestError as e:
        error_msg = f"Request error downloading from {request.url}: {e!s}"
        logger.error(error_msg)
        return DownloadFromUrlResponse(
            file_path=request.file_path,
            file_size_bytes=0,
            success=False,
            error_message=error_msg,
        )

    except OSError as e:
        error_msg = f"File system error writing to {request.file_path}: {e!s}"
        logger.error(error_msg)
        return DownloadFromUrlResponse(
            file_path=request.file_path,
            file_size_bytes=0,
            success=False,
            error_message=error_msg,
        )

    except Exception as e:
        error_msg = f"Unexpected error downloading from {request.url}: {e!s}"
        logger.error(error_msg)
        return DownloadFromUrlResponse(
            file_path=request.file_path,
            file_size_bytes=0,
            success=False,
            error_message=error_msg,
        )



# Generated by ariadne-codegen
# Source: ../../schema.graphql

from typing import Any, Optional
from uuid import UUID

from pydantic import Field

from .base_model import BaseModel
from .enums import (
    AgentSubtype,
    ChatMode,
    ModelName,
    ReasoningLevel,
    SandboxProvider,
    ToolPermissionStatus,
)


class AttachmentInput(BaseModel):
    file_attachment: Optional["FileAttachmentInput"] = Field(
        alias="fileAttachment", default=None
    )
    url_attachment: Optional["URLAttachmentInput"] = Field(
        alias="urlAttachment", default=None
    )
    table_schema_attachment: Optional["TableSchemaAttachmentInput"] = Field(
        alias="tableSchemaAttachment", default=None
    )
    prompt_attachment: Optional["PromptAttachmentInput"] = Field(
        alias="promptAttachment", default=None
    )
    sql_attachment: Optional["SQLAttachmentInput"] = Field(
        alias="sqlAttachment", default=None
    )
    skill_attachment: Optional["SkillAttachmentInput"] = Field(
        alias="skillAttachment", default=None
    )


class ChatConfig(BaseModel):
    chat_uuid: UUID = Field(alias="chatUuid")
    model: Optional[ModelName] = None
    require_confirmation: Optional[bool] = Field(
        alias="requireConfirmation", default=False
    )
    read_only: Optional[bool] = Field(alias="readOnly", default=False)
    depth_limit: int = Field(alias="depthLimit", default=6)
    reasoning_level: ReasoningLevel = Field(
        alias="reasoningLevel", default=ReasoningLevel.LOW
    )
    internal_settings: Optional[Any] = Field(alias="internalSettings", default=None)


class ChatInput(BaseModel):
    prompt: Optional["Prompt"] = None
    client_sourced_event_uuid: Optional[UUID] = Field(
        alias="clientSourcedEventUuid", default=None
    )
    permission_status: Optional[ToolPermissionStatus] = Field(
        alias="permissionStatus", default=None
    )
    table_filter_config_uuid: Optional[UUID] = Field(
        alias="tableFilterConfigUuid", default=None
    )
    planning_required: Optional[bool] = Field(alias="planningRequired", default=None)
    timezone: Optional[str] = None
    mcp_server_configs: Optional["McpServerConfigsInput"] = Field(
        alias="mcpServerConfigs", default=None
    )
    database_resource_configs: Optional[list["DatabaseResourceConfigInput"]] = Field(
        alias="databaseResourceConfigs", default=None
    )
    repository_resource_config: Optional["RepositoryResourceConfigInput"] = Field(
        alias="repositoryResourceConfig", default=None
    )


class ChatResourceConfigInput(BaseModel):
    mode: ChatMode
    repositories: Optional[list["RepositoryResourceConfigInput"]] = None
    databases: Optional[list["DatabaseResourceConfigInput"]] = None
    mcp_server_configs: Optional["McpServerConfigsInput"] = Field(
        alias="mcpServerConfigs", default=None
    )
    skill_configs: Optional["SkillConfigsInput"] = Field(
        alias="skillConfigs", default=None
    )
    sandbox_provider: Optional[SandboxProvider] = Field(
        alias="sandboxProvider", default=None
    )
    cloud_agent_subtype: Optional[AgentSubtype] = Field(
        alias="cloudAgentSubtype", default=None
    )


class DatabaseResourceConfigInput(BaseModel):
    database_config_uuid: UUID = Field(alias="databaseConfigUuid")


class FileAttachmentInput(BaseModel):
    file: "FileInput"
    content: str


class FileInput(BaseModel):
    file_path: str = Field(alias="filePath")
    working_directory: str = Field(alias="workingDirectory")


class McpServerConfigInput(BaseModel):
    server_name: str = Field(alias="serverName")
    enabled_tools: list[str] = Field(alias="enabledTools")


class McpServerConfigsInput(BaseModel):
    all_servers: Optional[bool] = Field(alias="allServers", default=None)
    server_configs: Optional[list["McpServerConfigInput"]] = Field(
        alias="serverConfigs", default=None
    )


class Prompt(BaseModel):
    message: str
    attachments: list["AttachmentInput"]


class PromptAttachmentInput(BaseModel):
    prompt_name: str = Field(alias="promptName")
    prompt_content: str = Field(alias="promptContent")


class RepositoryInput(BaseModel):
    org_name: str = Field(alias="orgName")
    repo_name: str = Field(alias="repoName")


class RepositoryResourceConfigInput(BaseModel):
    repository_uuid: UUID = Field(alias="repositoryUuid")
    is_primary: bool = Field(alias="isPrimary", default=False)


class SQLAttachmentInput(BaseModel):
    query_content: str = Field(alias="queryContent")


class SkillAttachmentInput(BaseModel):
    skill_name: str = Field(alias="skillName")


class SkillConfigInput(BaseModel):
    skill_id: str = Field(alias="skillId")


class SkillConfigsInput(BaseModel):
    skill_configs: list["SkillConfigInput"] = Field(alias="skillConfigs")


class TableSchemaAttachmentInput(BaseModel):
    table_name: str = Field(alias="tableName")
    table_schema: Any = Field(alias="tableSchema")


class URLAttachmentInput(BaseModel):
    url: str
    content: str


AttachmentInput.model_rebuild()
ChatInput.model_rebuild()
ChatResourceConfigInput.model_rebuild()
FileAttachmentInput.model_rebuild()
McpServerConfigsInput.model_rebuild()
Prompt.model_rebuild()
SkillConfigsInput.model_rebuild()



CREATE_CLOUD_CHAT_FROM_REPOSITORY_MUTATION = """
mutation CreateCloudChatFromRepository($repositoryUuid: UUID!, $provider: SandboxProvider) {
  createCloudChat(
    input: {repositoryUuid: $repositoryUuid, provider: $provider}
  ) {
    __typename
    ...on Chat {
      chatUuid
    }
    ...on UnauthenticatedError {
      message
    }
    ...on NotFoundError {
      resourceType
      resourceId
      message
    }
    ...on CloudSessionError {
      message
    }
  }
}
"""



"""Generalized truncation framework for tool results."""

import os
import tempfile
from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar, cast

from msgspec.structs import replace

from exponent.core.config import get_chat_artifacts_dir
from exponent.core.remote_execution.cli_rpc_types import (
    BashToolResult,
    ErrorToolResult,
    GlobToolResult,
    GrepToolResult,
    ReadToolResult,
    ToolResult,
    WriteToolResult,
)
from exponent.core.remote_execution.utils import truncate_output

DEFAULT_CHARACTER_LIMIT = 20_000
BASH_CHARACTER_LIMIT = 8_000
MAX_FILE_CHARS = 2_000_000  # 2M character limit for full output file
DEFAULT_LIST_ITEM_LIMIT = 1000
DEFAULT_LIST_PREVIEW_ITEMS = 10


def _write_full_output_to_file(output: str, chat_uuid: str) -> str | None:
    """Write full output to a temp file, truncated to last MAX_FILE_CHARS characters.

    Files are written to the per-chat artifacts directory (~/.indent/chats/{uuid}/).
    """
    try:
        if len(output) > MAX_FILE_CHARS:
            output = output[-MAX_FILE_CHARS:]

        artifacts_dir = get_chat_artifacts_dir(chat_uuid)
        os.makedirs(artifacts_dir, exist_ok=True)

        fd, path = tempfile.mkstemp(prefix="bash_output_", suffix=".txt", dir=artifacts_dir)
        try:
            os.write(fd, output.encode("utf-8", errors="replace"))
        finally:
            os.close(fd)
        return path
    except Exception:
        return None


_T = TypeVar("_T", bound=ToolResult)


class TruncationStrategy(ABC, Generic[_T]):
    @abstractmethod
    def should_truncate(self, result: _T) -> bool:
        """Return True if the result should be truncated."""

    @abstractmethod
    def truncate(self, result: _T, chat_uuid: str | None = None) -> _T:
        """Truncate the result and return the truncated version."""


class StringFieldTruncation(TruncationStrategy[_T]):
    def __init__(
        self,
        field_name: str,
        character_limit: int = DEFAULT_CHARACTER_LIMIT,
    ):
        self.field_name = field_name
        self.character_limit = character_limit

    def should_truncate(self, result: _T) -> bool:
        if hasattr(result, self.field_name):
            value = getattr(result, self.field_name)
            if isinstance(value, str):
                return len(value) > self.character_limit
        return False

    def truncate(self, result: _T, chat_uuid: str | None = None) -> _T:
        if not hasattr(result, self.field_name):
            return result

        value = getattr(result, self.field_name)
        if not isinstance(value, str):
            return result

        truncated_value, was_truncated = truncate_output(value, self.character_limit)

        updates: dict[str, Any] = {self.field_name: truncated_value}
        if hasattr(result, "truncated") and was_truncated:
            updates["truncated"] = True

        return replace(result, **updates)


class ListFieldTruncation(TruncationStrategy[_T]):
    def __init__(
        self,
        field_name: str,
        item_limit: int = DEFAULT_LIST_ITEM_LIMIT,
        preview_items: int = DEFAULT_LIST_PREVIEW_ITEMS,
    ):
        self.field_name = field_name
        self.item_limit = item_limit
        self.preview_items = preview_items

    def should_truncate(self, result: _T) -> bool:
        if hasattr(result, self.field_name):
            value = getattr(result, self.field_name)
            if isinstance(value, list):
                return len(value) > self.item_limit
        return False

    def truncate(self, result: _T, chat_uuid: str | None = None) -> _T:
        if not hasattr(result, self.field_name):
            return result

        value = getattr(result, self.field_name)
        if not isinstance(value, list):
            return result

        total_items = len(value)
        if total_items <= self.item_limit:
            return result

        truncated_count = max(0, total_items - 2 * self.preview_items)
        truncated_list = (
            value[: self.preview_items] + [f"... {truncated_count} items truncated ..."] + value[-self.preview_items :]
        )

        updates: dict[str, Any] = {self.field_name: truncated_list}
        if hasattr(result, "truncated"):
            updates["truncated"] = True

        return replace(result, **updates)


class CompositeTruncation(TruncationStrategy[_T]):
    def __init__(self, strategies: list[TruncationStrategy[_T]]):
        self.strategies = strategies

    def should_truncate(self, result: _T) -> bool:
        return any(strategy.should_truncate(result) for strategy in self.strategies)

    def truncate(self, result: _T, chat_uuid: str | None = None) -> _T:
        for strategy in self.strategies:
            if strategy.should_truncate(result):
                result = strategy.truncate(result, chat_uuid)
        return result


class TailTruncation(TruncationStrategy[_T]):
    """Truncation strategy that keeps the end of the output (tail) instead of the beginning.

    Writes full output to a temp file before truncating.
    """

    def __init__(
        self,
        field_name: str,
        character_limit: int = DEFAULT_CHARACTER_LIMIT,
    ):
        self.field_name = field_name
        self.character_limit = character_limit

    def should_truncate(self, result: _T) -> bool:
        if hasattr(result, self.field_name):
            value = getattr(result, self.field_name)
            if isinstance(value, str):
                return len(value) > self.character_limit
        return False

    def truncate(self, result: _T, chat_uuid: str | None = None) -> _T:
        if not hasattr(result, self.field_name):
            return result

        value = getattr(result, self.field_name)
        if not isinstance(value, str):
            return result

        file_path = _write_full_output_to_file(value, chat_uuid) if chat_uuid else None

        updates: dict[str, Any] = {}
        if file_path and hasattr(result, "output_file"):
            updates["output_file"] = file_path

        if len(value) <= self.character_limit:
            if updates:
                return replace(result, **updates)
            return result

        truncated_value = value[-self.character_limit :]

        newline_pos = truncated_value.find("\n")
        if newline_pos != -1 and newline_pos < 1000:
            truncated_value = truncated_value[newline_pos + 1 :]

        if file_path:
            truncation_msg = (
                f"[Truncated to last {self.character_limit} characters. Full output written to: {file_path}]\n"
            )
        else:
            truncation_msg = f"[Truncated to last {self.character_limit} characters.]\n"
        truncated_value = truncation_msg + truncated_value

        updates[self.field_name] = truncated_value
        if hasattr(result, "truncated"):
            updates["truncated"] = True

        return replace(result, **updates)


class NoOpTruncation(TruncationStrategy[_T]):
    def should_truncate(self, result: _T) -> bool:
        return False

    def truncate(self, result: _T, chat_uuid: str | None = None) -> _T:
        return result


class StringListTruncation(TruncationStrategy[_T]):
    """Truncation for lists of strings that limits both number of items and individual string length."""

    def __init__(
        self,
        field_name: str,
        max_items: int = DEFAULT_LIST_ITEM_LIMIT,
        preview_items: int = DEFAULT_LIST_PREVIEW_ITEMS,
        max_item_length: int = 1000,
    ):
        self.field_name = field_name
        self.max_items = max_items
        self.preview_items = preview_items
        self.max_item_length = max_item_length

    def should_truncate(self, result: _T) -> bool:
        if not hasattr(result, self.field_name):
            return False

        items = getattr(result, self.field_name)
        if not isinstance(items, list):
            return False

        # Check if we need to truncate number of items
        if len(items) > self.max_items:
            return True

        # Check if any individual item is too long
        for item in items:
            if isinstance(item, str) and len(item) > self.max_item_length:
                return True
            # Handle dict items (e.g., with metadata like file path and line number)
            elif isinstance(item, dict) and "content" in item:
                if len(item["content"]) > self.max_item_length:
                    return True

        return False

    def _truncate_item_content(self, item: str | dict[str, Any]) -> str | dict[str, Any]:
        """Truncate an individual item's content."""
        if isinstance(item, str):
            if len(item) <= self.max_item_length:
                return item
            # Truncate string item
            truncated, _ = truncate_output(item, self.max_item_length)
            return truncated
        elif isinstance(item, dict) and "content" in item:
            # Handle dict-style items (e.g., with metadata like file path and line number)
            if len(item["content"]) <= self.max_item_length:
                return item
            truncated_content, _ = truncate_output(item["content"], self.max_item_length)
            return {**item, "content": truncated_content}
        else:
            return item

    def truncate(self, result: _T, chat_uuid: str | None = None) -> _T:
        if not hasattr(result, self.field_name):
            return result

        items = getattr(result, self.field_name)
        if not isinstance(items, list):
            return result

        # First, truncate individual item contents
        truncated_items = [self._truncate_item_content(item) for item in items]

        # Then, limit the number of items if needed
        total_items = len(truncated_items)
        if total_items > self.max_items:
            truncated_count = max(0, total_items - 2 * self.preview_items)
            final_items = (
                truncated_items[: self.preview_items]
                + [f"... {truncated_count} items truncated ..."]
                + truncated_items[-self.preview_items :]
            )
        else:
            final_items = truncated_items

        updates: dict[str, Any] = {self.field_name: final_items}
        if hasattr(result, "truncated"):
            updates["truncated"] = True

        return replace(result, **updates)


TRUNCATION_REGISTRY: dict[type[ToolResult], TruncationStrategy[Any]] = {
    ReadToolResult: StringFieldTruncation("content"),
    WriteToolResult: StringFieldTruncation("message"),
    GrepToolResult: StringListTruncation("matches"),
    GlobToolResult: StringListTruncation("filenames", max_item_length=4096),
    BashToolResult: TailTruncation("shell_output", character_limit=BASH_CHARACTER_LIMIT),
}


T = TypeVar("T", bound=ToolResult)


def truncate_tool_result(result: T, chat_uuid: str | None = None) -> T:
    if isinstance(result, ErrorToolResult):
        return result

    result_type = type(result)
    if result_type in TRUNCATION_REGISTRY:
        strategy = TRUNCATION_REGISTRY[result_type]
        if isinstance(result, BashToolResult):
            return cast(T, strategy.truncate(result, chat_uuid))
        elif strategy.should_truncate(result):
            return cast(T, strategy.truncate(result, chat_uuid))

    return result



# Generated by ariadne-codegen
# Source: exponent/core/graphql/operations

from typing import Literal, Optional, Union
from uuid import UUID

from pydantic import Field

from .base_model import BaseModel


class StartChatTurn(BaseModel):
    start_chat_turn: Union[
        "StartChatTurnStartChatTurnUnauthenticatedError",
        "StartChatTurnStartChatTurnError",
        "StartChatTurnStartChatTurnNotFoundError",
        "StartChatTurnStartChatTurnInputValidationError",
        "StartChatTurnStartChatTurnPlanError",
        "StartChatTurnStartChatTurnChat",
    ] = Field(alias="startChatTurn", discriminator="typename__")


class StartChatTurnStartChatTurnUnauthenticatedError(BaseModel):
    typename__: Literal["UnauthenticatedError"] = Field(alias="__typename")
    message: str


class StartChatTurnStartChatTurnError(BaseModel):
    typename__: Literal["Error"] = Field(alias="__typename")


class StartChatTurnStartChatTurnNotFoundError(BaseModel):
    typename__: Literal["NotFoundError"] = Field(alias="__typename")
    resource_type: str = Field(alias="resourceType")
    resource_id: Optional[str] = Field(alias="resourceId")
    message: str


class StartChatTurnStartChatTurnInputValidationError(BaseModel):
    typename__: Literal["InputValidationError"] = Field(alias="__typename")


class StartChatTurnStartChatTurnPlanError(BaseModel):
    typename__: Literal["PlanError"] = Field(alias="__typename")


class StartChatTurnStartChatTurnChat(BaseModel):
    typename__: Literal["Chat"] = Field(alias="__typename")
    chat_uuid: UUID = Field(alias="chatUuid")


StartChatTurn.model_rebuild()



# Generated by ariadne-codegen

from .base_client import AsyncBaseClient
from .base_model import BaseModel, Upload
from .chats import (
    Chats,
    ChatsChatsChats,
    ChatsChatsChatsChats,
    ChatsChatsChatsChatsPrStatusNotFoundError,
    ChatsChatsChatsChatsPrStatusPRInfo,
    ChatsChatsUnauthenticatedError,
)
from .client import IndentGraphQLClient
from .create_cloud_chat_from_repository import (
    CreateCloudChatFromRepository,
    CreateCloudChatFromRepositoryCreateChatWithResourceConfigChat,
    CreateCloudChatFromRepositoryCreateChatWithResourceConfigInputValidationError,
    CreateCloudChatFromRepositoryCreateChatWithResourceConfigNotFoundError,
    CreateCloudChatFromRepositoryCreateChatWithResourceConfigPlanError,
    CreateCloudChatFromRepositoryCreateChatWithResourceConfigUnauthenticatedError,
)
from .enable_cloud_repository import (
    EnableCloudRepository,
    EnableCloudRepositoryEnableCloudRepositoryCloudSessionError,
    EnableCloudRepositoryEnableCloudRepositoryEnableCloudRepositoriesResult,
    EnableCloudRepositoryEnableCloudRepositoryEnableCloudRepositoriesResultResults,
    EnableCloudRepositoryEnableCloudRepositoryEnableCloudRepositoriesResultResultsImages,
    EnableCloudRepositoryEnableCloudRepositoryInputValidationError,
    EnableCloudRepositoryEnableCloudRepositoryNotFoundError,
    EnableCloudRepositoryEnableCloudRepositoryUnauthenticatedError,
)
from .enums import (
    AgentSubtype,
    ChatMode,
    ModelName,
    PRStatus,
    ReasoningLevel,
    SandboxProvider,
    ToolPermissionStatus,
)
from .github_repositories import (
    GithubRepositories,
    GithubRepositoriesGithubRepositories,
    GithubRepositoriesGithubRepositoriesRepositories,
)
from .halt_chat_stream import (
    HaltChatStream,
    HaltChatStreamHaltChatStreamChat,
    HaltChatStreamHaltChatStreamNotFoundError,
    HaltChatStreamHaltChatStreamRemoteExecutionError,
    HaltChatStreamHaltChatStreamUnauthenticatedError,
)
from .input_types import (
    AttachmentInput,
    ChatConfig,
    ChatInput,
    ChatResourceConfigInput,
    DatabaseResourceConfigInput,
    FileAttachmentInput,
    FileInput,
    McpServerConfigInput,
    McpServerConfigsInput,
    Prompt,
    PromptAttachmentInput,
    RepositoryInput,
    RepositoryResourceConfigInput,
    SkillAttachmentInput,
    SkillConfigInput,
    SkillConfigsInput,
    SQLAttachmentInput,
    TableSchemaAttachmentInput,
    URLAttachmentInput,
)
from .rebuild_cloud_repository import (
    RebuildCloudRepository,
    RebuildCloudRepositoryRebuildCloudRepositoryCloudSessionError,
    RebuildCloudRepositoryRebuildCloudRepositoryContainerImages,
    RebuildCloudRepositoryRebuildCloudRepositoryContainerImagesImages,
    RebuildCloudRepositoryRebuildCloudRepositoryInputValidationError,
    RebuildCloudRepositoryRebuildCloudRepositoryNotFoundError,
    RebuildCloudRepositoryRebuildCloudRepositoryUnauthenticatedError,
)
from .refresh_api_key import (
    RefreshApiKey,
    RefreshApiKeyRefreshApiKeyUnauthenticatedError,
    RefreshApiKeyRefreshApiKeyUser,
)
from .report_sandbox_info import (
    ReportSandboxInfo,
    ReportSandboxInfoReportSandboxInfoSandboxInfoResponse,
    ReportSandboxInfoReportSandboxInfoUnauthenticatedError,
)
from .set_login_complete import (
    SetLoginComplete,
    SetLoginCompleteSetLoginCompleteUnauthenticatedError,
    SetLoginCompleteSetLoginCompleteUser,
)
from .start_chat_turn import (
    StartChatTurn,
    StartChatTurnStartChatTurnChat,
    StartChatTurnStartChatTurnError,
    StartChatTurnStartChatTurnInputValidationError,
    StartChatTurnStartChatTurnNotFoundError,
    StartChatTurnStartChatTurnPlanError,
    StartChatTurnStartChatTurnUnauthenticatedError,
)

__all__ = [
    "AgentSubtype",
    "AsyncBaseClient",
    "AttachmentInput",
    "BaseModel",
    "ChatConfig",
    "ChatInput",
    "ChatMode",
    "ChatResourceConfigInput",
    "Chats",
    "ChatsChatsChats",
    "ChatsChatsChatsChats",
    "ChatsChatsChatsChatsPrStatusNotFoundError",
    "ChatsChatsChatsChatsPrStatusPRInfo",
    "ChatsChatsUnauthenticatedError",
    "CreateCloudChatFromRepository",
    "CreateCloudChatFromRepositoryCreateChatWithResourceConfigChat",
    "CreateCloudChatFromRepositoryCreateChatWithResourceConfigInputValidationError",
    "CreateCloudChatFromRepositoryCreateChatWithResourceConfigNotFoundError",
    "CreateCloudChatFromRepositoryCreateChatWithResourceConfigPlanError",
    "CreateCloudChatFromRepositoryCreateChatWithResourceConfigUnauthenticatedError",
    "DatabaseResourceConfigInput",
    "EnableCloudRepository",
    "EnableCloudRepositoryEnableCloudRepositoryCloudSessionError",
    "EnableCloudRepositoryEnableCloudRepositoryEnableCloudRepositoriesResult",
    "EnableCloudRepositoryEnableCloudRepositoryEnableCloudRepositoriesResultResults",
    "EnableCloudRepositoryEnableCloudRepositoryEnableCloudRepositoriesResultResultsImages",
    "EnableCloudRepositoryEnableCloudRepositoryInputValidationError",
    "EnableCloudRepositoryEnableCloudRepositoryNotFoundError",
    "EnableCloudRepositoryEnableCloudRepositoryUnauthenticatedError",
    "FileAttachmentInput",
    "FileInput",
    "GithubRepositories",
    "GithubRepositoriesGithubRepositories",
    "GithubRepositoriesGithubRepositoriesRepositories",
    "HaltChatStream",
    "HaltChatStreamHaltChatStreamChat",
    "HaltChatStreamHaltChatStreamNotFoundError",
    "HaltChatStreamHaltChatStreamRemoteExecutionError",
    "HaltChatStreamHaltChatStreamUnauthenticatedError",
    "IndentGraphQLClient",
    "McpServerConfigInput",
    "McpServerConfigsInput",
    "ModelName",
    "PRStatus",
    "Prompt",
    "PromptAttachmentInput",
    "ReasoningLevel",
    "RebuildCloudRepository",
    "RebuildCloudRepositoryRebuildCloudRepositoryCloudSessionError",
    "RebuildCloudRepositoryRebuildCloudRepositoryContainerImages",
    "RebuildCloudRepositoryRebuildCloudRepositoryContainerImagesImages",
    "RebuildCloudRepositoryRebuildCloudRepositoryInputValidationError",
    "RebuildCloudRepositoryRebuildCloudRepositoryNotFoundError",
    "RebuildCloudRepositoryRebuildCloudRepositoryUnauthenticatedError",
    "RefreshApiKey",
    "RefreshApiKeyRefreshApiKeyUnauthenticatedError",
    "RefreshApiKeyRefreshApiKeyUser",
    "ReportSandboxInfo",
    "ReportSandboxInfoReportSandboxInfoSandboxInfoResponse",
    "ReportSandboxInfoReportSandboxInfoUnauthenticatedError",
    "RepositoryInput",
    "RepositoryResourceConfigInput",
    "SQLAttachmentInput",
    "SandboxProvider",
    "SetLoginComplete",
    "SetLoginCompleteSetLoginCompleteUnauthenticatedError",
    "SetLoginCompleteSetLoginCompleteUser",
    "SkillAttachmentInput",
    "SkillConfigInput",
    "SkillConfigsInput",
    "StartChatTurn",
    "StartChatTurnStartChatTurnChat",
    "StartChatTurnStartChatTurnError",
    "StartChatTurnStartChatTurnInputValidationError",
    "StartChatTurnStartChatTurnNotFoundError",
    "StartChatTurnStartChatTurnPlanError",
    "StartChatTurnStartChatTurnUnauthenticatedError",
    "TableSchemaAttachmentInput",
    "ToolPermissionStatus",
    "URLAttachmentInput",
    "Upload",
]



import asyncio
import logging
import os
import shutil
import tempfile
import urllib.request
import uuid
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from time import time
from typing import TYPE_CHECKING

import aiohttp
from anyio import Path as AsyncPath

from exponent.core.config import get_chat_artifacts_dir
from exponent.core.remote_execution import files
from exponent.core.remote_execution.cli_rpc_types import (
    BashToolInput,
    BashToolResult,
    DownloadArtifactToolInput,
    DownloadArtifactToolResult,
    EditToolInput,
    EditToolResult,
    ErrorToolResult,
    GlobToolInput,
    GlobToolResult,
    GrepToolInput,
    GrepToolResult,
    ReadToolArtifactResult,
    ReadToolInput,
    ReadToolResult,
    StoreArtifactToolInput,
    StoreArtifactToolResult,
    StreamingCodeExecutionRequest,
    StreamingCodeExecutionResponse,
    ToolInputType,
    ToolResultType,
    UploadArtifactToolInput,
    UploadArtifactToolResult,
    WriteToolInput,
    WriteToolResult,
)
from exponent.core.remote_execution.code_execution import (
    execute_code_streaming,
)
from exponent.core.remote_execution.default_env import get_process_env
from exponent.core.remote_execution.file_write import execute_full_file_rewrite
from exponent.core.remote_execution.languages.shell_streaming import (
    get_rc_file_source_command,
)
from exponent.core.remote_execution.truncation import truncate_tool_result
from exponent.core.remote_execution.utils import (
    assert_unreachable,
    safe_get_file_metadata,
    safe_read_file,
)

if TYPE_CHECKING:
    from exponent.core.remote_execution.client import RemoteExecutionClient

logger = logging.getLogger(__name__)


@dataclass
class BackgroundBashResult:
    """Result from spawning a background bash process.

    Contains both the tool result to return to the caller and the process
    object for tracking completion.
    """

    result: BashToolResult
    process: asyncio.subprocess.Process


async def execute_tool(
    tool_input: ToolInputType,
    working_directory: str,
    client: "RemoteExecutionClient",
) -> ToolResultType:
    if isinstance(tool_input, ReadToolInput):
        return await execute_read_file(tool_input, working_directory, client)
    elif isinstance(tool_input, WriteToolInput):
        return await execute_write_file(tool_input, working_directory)
    elif isinstance(tool_input, GlobToolInput):
        return await execute_glob_files(tool_input, working_directory)
    elif isinstance(tool_input, GrepToolInput):
        return await execute_grep_files(tool_input, working_directory)
    elif isinstance(tool_input, EditToolInput):
        return await execute_edit_file(tool_input, working_directory)
    elif isinstance(tool_input, DownloadArtifactToolInput):
        return await execute_download_artifact(tool_input, working_directory)
    elif isinstance(tool_input, UploadArtifactToolInput):
        return await execute_upload_artifact(tool_input, working_directory)
    elif isinstance(tool_input, StoreArtifactToolInput):
        return await execute_store_artifact(tool_input, working_directory, client)
    elif isinstance(tool_input, BashToolInput):
        raise ValueError("Bash tool input should be handled by execute_bash_tool")
    else:
        assert_unreachable(tool_input)


def truncate_result[T: ToolResultType](tool_result: T, chat_uuid: str) -> T:
    return truncate_tool_result(tool_result, chat_uuid)


def is_image_file(file_path: str) -> tuple[bool, str | None]:
    ext = Path(file_path).suffix.lower()
    if ext == ".png":
        return (True, "image/png")
    elif ext in [".jpg", ".jpeg"]:
        return (True, "image/jpeg")
    return (False, None)


async def execute_read_file(  # noqa: PLR0911
    tool_input: ReadToolInput,
    working_directory: str,
    client: "RemoteExecutionClient",
) -> ReadToolResult | ErrorToolResult:
    # Validate absolute path requirement
    if not tool_input.file_path.startswith("/"):
        return ErrorToolResult(error_message=f"File path must be absolute, got relative path: {tool_input.file_path}")

    # Validate offset and limit
    offset = tool_input.offset if tool_input.offset is not None else 0
    limit = tool_input.limit if tool_input.limit is not None else 2000

    if limit <= 0:
        return ErrorToolResult(error_message=f"Limit must be positive, got: {limit}")

    file = AsyncPath(working_directory, tool_input.file_path)

    # Check if this is an image file
    is_image, media_type = is_image_file(tool_input.file_path)
    if is_image and media_type:
        try:
            file_name = Path(tool_input.file_path).name
            s3_key = f"images/{uuid.uuid4()}/{file_name}"

            upload_response = await client.request_upload_url(s3_key, media_type)

            f = await file.open("rb")
            async with f:
                file_data = await f.read()

                def _upload() -> int:
                    req = urllib.request.Request(
                        upload_response.upload_url,
                        data=file_data,
                        headers={"Content-Type": media_type},
                        method="PUT",
                    )
                    with urllib.request.urlopen(req) as resp:
                        status: int = resp.status
                        return status

                status = await asyncio.to_thread(_upload)
                if status != 200:
                    raise RuntimeError(f"Upload failed with status {status}")

            return ReadToolResult(
                artifact=ReadToolArtifactResult(
                    s3_uri=upload_response.s3_uri,
                    file_path=tool_input.file_path,
                    media_type=media_type,
                )
            )
        except Exception as e:
            return ErrorToolResult(error_message=f"Failed to upload image to S3: {e!s}")

    try:
        exists = await file.exists()
    except (OSError, PermissionError) as e:
        return ErrorToolResult(error_message=f"Cannot access file: {e!s}")

    if not exists:
        return ErrorToolResult(
            error_message="File not found",
        )

    try:
        if await file.is_dir():
            return ErrorToolResult(
                error_message=f"{await file.absolute()} is a directory",
            )
    except (OSError, PermissionError) as e:
        return ErrorToolResult(error_message=f"Cannot check file type: {e!s}")

    try:
        content = await safe_read_file(file)
    except PermissionError:
        return ErrorToolResult(error_message=f"Permission denied: cannot read {tool_input.file_path}")
    except UnicodeDecodeError:
        return ErrorToolResult(error_message="File appears to be binary or has invalid text encoding")
    except Exception as e:
        return ErrorToolResult(error_message=f"Error reading file: {e!s}")

    metadata = await safe_get_file_metadata(file)

    # Handle empty files
    if not content:
        return ReadToolResult(
            content="",
            num_lines=0,
            start_line=0,
            total_lines=0,
            metadata=metadata,
        )

    content_lines = content.splitlines(keepends=True)
    total_lines = len(content_lines)

    # Handle offset beyond file length for positive offsets
    if offset >= 0 and offset >= total_lines:
        return ReadToolResult(
            content="",
            num_lines=0,
            start_line=offset,
            total_lines=total_lines,
            metadata=metadata,
        )

    # Use Python's native slicing - it handles negative offsets naturally
    # Handle the case where offset + limit < 0 (can't mix negative and non-negative indices)
    if offset < 0 and offset + limit < 0:
        # Both start and end are negative, use negative end index
        end_index = offset + limit
    elif offset < 0 and offset + limit >= 0:
        # Start is negative but end would be positive/zero, slice to end
        end_index = None
    else:
        # Normal case: both indices are non-negative
        end_index = offset + limit

    content_lines = content_lines[offset:end_index]

    # Calculate the actual start line for the result
    if offset < 0:
        # For negative offsets, calculate where we actually started
        actual_start_line = max(0, total_lines + offset)
    else:
        actual_start_line = offset

    # Apply character-level truncation at line boundaries to ensure consistency
    # This ensures the content field and num_lines field remain in sync
    # This is the primary truncation point for ReadToolResult - truncation.py
    # also has a fallback but this handles line-count synchronization
    CHARACTER_LIMIT = 20_000  # Match DEFAULT_CHARACTER_LIMIT in truncation.py

    # Join lines and check total size
    final_content = "".join(content_lines)

    if len(final_content) > CHARACTER_LIMIT:
        # Truncate at line boundaries to stay under the limit
        truncated_lines: list[str] = []
        current_size = 0
        truncation_message = "\n[Content truncated due to size limit]"
        truncation_size = len(truncation_message)
        lines_included = 0

        for line in content_lines:
            # Check if adding this line would exceed the limit (accounting for truncation message)
            if current_size + len(line) + truncation_size > CHARACTER_LIMIT:
                final_content = "".join(truncated_lines) + truncation_message
                break
            truncated_lines.append(line)
            current_size += len(line)
            lines_included += 1
        else:
            # All lines fit (shouldn't happen if we got here, but be safe)
            final_content = "".join(truncated_lines)
            lines_included = len(content_lines)

        num_lines = lines_included
    else:
        num_lines = len(content_lines)

    return ReadToolResult(
        content=final_content,
        num_lines=num_lines,
        start_line=actual_start_line,
        total_lines=total_lines,
        metadata=metadata,
    )


async def execute_write_file(tool_input: WriteToolInput, working_directory: str) -> WriteToolResult:
    file_path = tool_input.file_path
    path = Path(working_directory, file_path)
    result = await execute_full_file_rewrite(path, tool_input.content, working_directory)
    return WriteToolResult(message=result)


async def execute_edit_file(  # noqa: PLR0911
    tool_input: EditToolInput, working_directory: str
) -> EditToolResult | ErrorToolResult:
    # Validate absolute path requirement
    if not tool_input.file_path.startswith("/"):
        return ErrorToolResult(error_message=f"File path must be absolute, got relative path: {tool_input.file_path}")

    file = AsyncPath(working_directory, tool_input.file_path)

    try:
        exists = await file.exists()
    except (OSError, PermissionError) as e:
        return ErrorToolResult(error_message=f"Cannot access file: {e!s}")

    if not exists:
        return ErrorToolResult(error_message="File not found")

    if tool_input.last_known_modified_timestamp is not None:
        metadata = await safe_get_file_metadata(file)
        if metadata is not None and metadata.modified_timestamp > tool_input.last_known_modified_timestamp:
            return ErrorToolResult(error_message="File has been modified since last read/write")

    try:
        if await file.is_dir():
            return ErrorToolResult(error_message=f"{await file.absolute()} is a directory")
    except (OSError, PermissionError) as e:
        return ErrorToolResult(error_message=f"Cannot check file type: {e!s}")

    try:
        # Read the entire file without truncation limits
        content = await safe_read_file(file)
    except PermissionError:
        return ErrorToolResult(error_message=f"Permission denied: cannot read {tool_input.file_path}")
    except UnicodeDecodeError:
        return ErrorToolResult(error_message="File appears to be binary or has invalid text encoding")
    except Exception as e:
        return ErrorToolResult(error_message=f"Error reading file: {e!s}")

    # Check if search text exists
    if tool_input.old_string not in content:
        return ErrorToolResult(error_message=f"Search text not found in {tool_input.file_path}")

    # Handle `old_string_end` parameter for range-based replacement (inclusive)
    if tool_input.old_string_end is not None:
        if tool_input.replace_all:
            return ErrorToolResult(error_message="Cannot use 'old_string_end' with 'replace_all=True'")

        # Both old_string and old_string_end must be unique
        old_string_occurrences = content.count(tool_input.old_string)
        if old_string_occurrences > 1:
            return ErrorToolResult(
                error_message=f"String '{tool_input.old_string}' appears {old_string_occurrences} times in file. Use a more specific start marker."
            )

        end_occurrences = content.count(tool_input.old_string_end)
        if end_occurrences > 1:
            return ErrorToolResult(
                error_message="Multiple matches for 'old_string_end'. Use a more specific end marker."
            )

        start_idx = content.find(tool_input.old_string)
        if start_idx == -1:
            return ErrorToolResult(error_message=f"'old_string' text not found in {tool_input.file_path}")

        search_after_start = start_idx + len(tool_input.old_string)
        until_idx = content.find(tool_input.old_string_end, search_after_start)
        if until_idx == -1:
            return ErrorToolResult(
                error_message=f"'old_string_end' text not found after 'old_string' in {tool_input.file_path}"
            )

        end_idx = until_idx + len(tool_input.old_string_end)

        old_text = content[start_idx:end_idx]
        if old_text == tool_input.new_string:
            return ErrorToolResult(error_message="Old string and new string are identical")

        new_content = content[:start_idx] + tool_input.new_string + content[end_idx:]
        replaced_old_lines = old_text.count("\n") + 1
        replaced_new_lines = tool_input.new_string.count("\n") + 1
    else:
        # Standard replacement logic
        if tool_input.old_string == tool_input.new_string:
            return ErrorToolResult(error_message="Old string and new string are identical")

        if not tool_input.replace_all:
            occurrences = content.count(tool_input.old_string)
            if occurrences > 1:
                return ErrorToolResult(
                    error_message=f"String '{tool_input.old_string}' appears {occurrences} times in file. Use a larger context or replace_all=True"
                )

        if tool_input.replace_all:
            new_content = content.replace(tool_input.old_string, tool_input.new_string)
        else:
            new_content = content.replace(tool_input.old_string, tool_input.new_string, 1)
        replaced_old_lines = tool_input.old_string.count("\n") + 1
        replaced_new_lines = tool_input.new_string.count("\n") + 1

    # Write back to file
    try:
        path = Path(working_directory, tool_input.file_path)
        await execute_full_file_rewrite(path, new_content, working_directory)
        return EditToolResult(
            message=f"Replaced {replaced_old_lines} lines with {replaced_new_lines} lines in {tool_input.file_path}",
            metadata=await safe_get_file_metadata(path),
        )
    except Exception as e:
        return ErrorToolResult(error_message=f"Error writing file: {e!s}")


async def execute_glob_files(tool_input: GlobToolInput, working_directory: str) -> GlobToolResult:
    # async timer
    start_time = time()
    results = await files.glob(
        path=working_directory if tool_input.path is None else tool_input.path,
        glob_pattern=tool_input.pattern,
    )
    duration_ms = int((time() - start_time) * 1000)
    return GlobToolResult(
        filenames=results,
        duration_ms=duration_ms,
        num_files=len(results),
        truncated=len(results) >= files.GLOB_MAX_COUNT,
    )


async def execute_grep_files(tool_input: GrepToolInput, working_directory: str) -> GrepToolResult | ErrorToolResult:
    return await files.search_files(
        path_str=working_directory if tool_input.path is None else tool_input.path,
        file_pattern=tool_input.include,
        regex=tool_input.pattern,
        working_directory=working_directory,
        multiline=tool_input.multiline,
    )


async def execute_bash_tool(
    tool_input: BashToolInput,
    working_directory: str,
    should_halt: Callable[[], bool],
    chat_uuid: str,
) -> BashToolResult | BackgroundBashResult:
    """Execute a bash command.

    For background commands, returns BackgroundBashResult containing both the result
    and the process object for tracking. For foreground commands, returns BashToolResult.
    """
    if tool_input.background:
        return await execute_bash_tool_background(tool_input, working_directory, chat_uuid)

    start_time = time()
    result = None
    async for result in execute_code_streaming(
        StreamingCodeExecutionRequest(
            language="shell",
            content=tool_input.command,
            timeout=120 if tool_input.timeout is None else tool_input.timeout,
            correlation_id=str(uuid.uuid4()),
        ),
        working_directory=working_directory,
        should_halt=should_halt,
        chat_uuid=chat_uuid,
    ):
        pass

    assert isinstance(result, StreamingCodeExecutionResponse)

    return BashToolResult(
        shell_output=result.content,
        exit_code=result.exit_code,
        duration_ms=int((time() - start_time) * 1000),
        timed_out=result.cancelled_for_timeout,
        stopped_by_user=result.halted,
    )


async def execute_bash_tool_background(
    tool_input: BashToolInput, working_directory: str, chat_uuid: str
) -> BackgroundBashResult:
    """Execute a bash command in the background, returning immediately with PID and output file.

    The command is spawned as a detached process with stdout/stderr redirected to a temp file.
    Output is streamed to the file in real-time, so it can be read while the process runs.

    Log files are written to the per-chat artifacts directory (~/.indent/chats/{uuid}/).

    Returns a BackgroundBashResult containing both the tool result and the process object
    for tracking completion.
    """
    start_time = time()

    artifacts_dir = get_chat_artifacts_dir(chat_uuid)
    os.makedirs(artifacts_dir, exist_ok=True)

    fd, output_file = tempfile.mkstemp(prefix="bg_", suffix=".log", dir=artifacts_dir)
    os.close(fd)

    shell_path = os.environ.get("SHELL") or shutil.which("bash") or shutil.which("sh")
    if not shell_path:
        shell_path = "/bin/sh"

    rc_source_cmd = get_rc_file_source_command(shell_path)
    full_command = f"{rc_source_cmd}{tool_input.command}"

    output_fd = os.open(output_file, os.O_WRONLY | os.O_CREAT | os.O_TRUNC)

    try:
        process = await asyncio.create_subprocess_exec(
            shell_path,
            "-l",
            "-c",
            full_command,
            stdin=asyncio.subprocess.DEVNULL,
            stdout=output_fd,
            stderr=output_fd,
            cwd=working_directory,
            env=get_process_env(None),
            start_new_session=True,
        )
    finally:
        os.close(output_fd)

    result = BashToolResult(
        shell_output=f"Background process started. Output is being written to: {output_file}",
        duration_ms=int((time() - start_time) * 1000),
        exit_code=None,
        timed_out=False,
        stopped_by_user=False,
        pid=process.pid,
        output_file=output_file,
    )

    return BackgroundBashResult(result=result, process=process)


async def _download_from_presigned_url(presigned_url: str, dest_path: Path) -> bytes:
    """Download content from a presigned URL and write to dest_path.

    Creates parent directories if needed. Returns the downloaded bytes.
    Raises aiohttp.ClientResponseError on HTTP errors.
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(presigned_url) as response:
            if response.status != 200:
                error_text = await response.text()
                raise aiohttp.ClientResponseError(
                    response.request_info,
                    response.history,
                    status=response.status,
                    message=error_text,
                )
            content = await response.read()
            dest_path.parent.mkdir(parents=True, exist_ok=True)
            dest_path.write_bytes(content)
            return content


async def execute_download_artifact(
    tool_input: DownloadArtifactToolInput, working_directory: str
) -> DownloadArtifactToolResult | ErrorToolResult:
    """Download an artifact from S3 using a pre-signed URL."""

    if not tool_input.file_path.startswith("/"):
        return ErrorToolResult(error_message=f"File path must be absolute, got relative path: {tool_input.file_path}")

    file_path = Path(tool_input.file_path)
    if file_path.exists() and not tool_input.overwrite:
        return ErrorToolResult(
            error_message=f"File already exists: {tool_input.file_path}. Set overwrite=True to replace it."
        )

    try:
        content = await _download_from_presigned_url(tool_input.presigned_url, file_path)
        file_size = len(content)

        content_preview = None
        num_lines = None
        total_lines = None
        truncated = False

        try:
            text_content = content.decode("utf-8")
            lines = text_content.splitlines()
            total_lines = len(lines)

            preview_limit = 50
            if len(lines) > preview_limit:
                preview_lines = lines[:preview_limit]
                truncated = True
                num_lines = preview_limit
            else:
                preview_lines = lines
                num_lines = len(lines)

            content_preview = "\n".join(preview_lines)
        except UnicodeDecodeError:
            pass

        return DownloadArtifactToolResult(
            file_path=tool_input.file_path,
            artifact_id=tool_input.artifact_id,
            file_size_bytes=file_size,
            content_preview=content_preview,
            num_lines=num_lines,
            total_lines=total_lines,
            truncated=truncated,
        )
    except aiohttp.ClientResponseError as e:
        return ErrorToolResult(error_message=f"Failed to download artifact: HTTP {e.status} - {e.message}")
    except Exception as e:
        logger.exception("Failed to download artifact")
        return ErrorToolResult(error_message=f"Failed to download artifact: {e!s}")


async def execute_upload_artifact(
    tool_input: UploadArtifactToolInput, working_directory: str
) -> UploadArtifactToolResult | ErrorToolResult:
    """Upload an artifact to S3 using a pre-signed URL."""

    if not tool_input.file_path.startswith("/"):
        return ErrorToolResult(error_message=f"File path must be absolute, got relative path: {tool_input.file_path}")

    file_path = Path(tool_input.file_path)
    if not file_path.exists():
        return ErrorToolResult(error_message=f"File not found: {tool_input.file_path}")

    if not file_path.is_file():
        return ErrorToolResult(error_message=f"Path is not a file: {tool_input.file_path}")

    try:
        content = file_path.read_bytes()
        file_size = len(content)

        async with aiohttp.ClientSession() as session:
            headers = {"Content-Type": tool_input.content_type}
            async with session.put(tool_input.presigned_url, data=content, headers=headers) as response:
                if response.status not in (200, 204):
                    error_text = await response.text()
                    return ErrorToolResult(
                        error_message=f"Failed to upload artifact: HTTP {response.status} - {error_text}"
                    )

        return UploadArtifactToolResult(
            artifact_id=tool_input.artifact_id,
            file_size_bytes=file_size,
            content_type=tool_input.content_type,
        )

    except Exception as e:
        logger.exception("Failed to upload artifact")
        return ErrorToolResult(error_message=f"Failed to upload artifact: {e!s}")


async def execute_store_artifact(
    tool_input: StoreArtifactToolInput,
    working_directory: str,
    client: "RemoteExecutionClient",
) -> StoreArtifactToolResult | ErrorToolResult:
    """Store an artifact to the CLI's config directory.

    Downloads from a presigned URL and stores in ~/.indent/chats/{chat_uuid}/{filename}.
    The CLI resolves the path using its own INDENT_HOME configuration.
    """
    if "/" in tool_input.filename or "\\" in tool_input.filename or ".." in tool_input.filename:
        return ErrorToolResult(
            error_message=f"Invalid filename (must not contain path separators): {tool_input.filename}"
        )

    chat_uuid = client.chat_uuid
    artifacts_dir = get_chat_artifacts_dir(chat_uuid)
    local_path = Path(artifacts_dir) / tool_input.filename

    try:
        content = await _download_from_presigned_url(tool_input.presigned_url, local_path)
        return StoreArtifactToolResult(
            local_path=str(local_path),
            file_size_bytes=len(content),
        )
    except aiohttp.ClientResponseError as e:
        return ErrorToolResult(error_message=f"Failed to download artifact: HTTP {e.status} - {e.message}")
    except Exception as e:
        logger.exception("Failed to store artifact")
        return ErrorToolResult(error_message=f"Failed to store artifact: {e!s}")



from __future__ import annotations

import sys
from dataclasses import dataclass
from pathlib import Path

import anyio
from anyio import Path as AsyncPath

BEGIN_PATCH = "*** Begin Patch"
END_PATCH = "*** End Patch"
ADD_FILE = "*** Add File: "
DELETE_FILE = "*** Delete File: "
UPDATE_FILE = "*** Update File: "
MOVE_TO = "*** Move to: "
EOF_MARKER = "*** End of File"
CHANGE_CONTEXT = "@@ "
EMPTY_CHANGE_CONTEXT = "@@"


class PatchError(Exception):
    pass


class InvalidPatchError(PatchError):
    pass


class InvalidHunkError(PatchError):
    def __init__(self, message: str, line_number: int) -> None:
        super().__init__(message)
        self.line_number = line_number


@dataclass
class UpdateFileChunk:
    change_context: str | None
    old_lines: list[str]
    new_lines: list[str]
    is_end_of_file: bool = False


@dataclass
class AddFile:
    path: str
    contents: str


@dataclass
class DeleteFile:
    path: str


@dataclass
class UpdateFile:
    path: str
    move_path: str | None
    chunks: list[UpdateFileChunk]


Hunk = AddFile | DeleteFile | UpdateFile


@dataclass
class AffectedPaths:
    added: list[Path]
    modified: list[Path]
    deleted: list[Path]


def parse_patch(patch: str) -> list[Hunk]:
    lines = _normalize_lines(patch)
    lines = _check_patch_boundaries_lenient(lines)
    if len(lines) < 2:
        raise InvalidPatchError("The first line of the patch must be '*** Begin Patch'")
    hunks: list[Hunk] = []
    remaining = lines[1:-1]
    line_number = 2
    index = 0
    while index < len(remaining):
        hunk, consumed = _parse_one_hunk(remaining[index:], line_number)
        hunks.append(hunk)
        index += consumed
        line_number += consumed
    return hunks


def _normalize_lines(patch: str) -> list[str]:
    return [line.rstrip("\r") for line in patch.strip().split("\n")]


def _check_patch_boundaries_lenient(lines: list[str]) -> list[str]:
    try:
        _check_patch_boundaries_strict(lines)
        return lines
    except InvalidPatchError as err:
        if not lines:
            raise err
        first = lines[0]
        last = lines[-1]
        if first in ("<<EOF", "<<'EOF'", '<<"EOF"') and last.endswith("EOF") and len(lines) >= 4:
            inner = lines[1:-1]
            _check_patch_boundaries_strict(inner)
            return inner
        raise err


def _check_patch_boundaries_strict(lines: list[str]) -> None:
    if not lines:
        raise InvalidPatchError("The first line of the patch must be '*** Begin Patch'")
    first = lines[0].strip()
    last = lines[-1].strip()
    if first != BEGIN_PATCH:
        raise InvalidPatchError("The first line of the patch must be '*** Begin Patch'")
    if last != END_PATCH:
        raise InvalidPatchError("The last line of the patch must be '*** End Patch'")


def _parse_one_hunk(lines: list[str], line_number: int) -> tuple[Hunk, int]:
    first_line = lines[0].strip()
    if first_line.startswith(ADD_FILE):
        path = first_line[len(ADD_FILE) :]
        contents = ""
        parsed_lines = 1
        for add_line in lines[1:]:
            if add_line.startswith("+"):
                contents += add_line[1:] + "\n"
                parsed_lines += 1
            else:
                break
        return AddFile(path=path, contents=contents), parsed_lines
    if first_line.startswith(DELETE_FILE):
        path = first_line[len(DELETE_FILE) :]
        return DeleteFile(path=path), 1
    if first_line.startswith(UPDATE_FILE):
        path = first_line[len(UPDATE_FILE) :]
        remaining = lines[1:]
        parsed_lines = 1
        move_path = None
        if remaining and remaining[0].startswith(MOVE_TO):
            move_path = remaining[0][len(MOVE_TO) :]
            remaining = remaining[1:]
            parsed_lines += 1

        chunks: list[UpdateFileChunk] = []
        while remaining:
            if remaining[0].strip() == "":
                parsed_lines += 1
                remaining = remaining[1:]
                continue
            if remaining[0].startswith("***"):
                break
            chunk, chunk_lines = _parse_update_file_chunk(remaining, line_number + parsed_lines, len(chunks) == 0)
            chunks.append(chunk)
            parsed_lines += chunk_lines
            remaining = remaining[chunk_lines:]

        if not chunks:
            raise InvalidHunkError(f"Update file hunk for path '{path}' is empty", line_number)
        return UpdateFile(path=path, move_path=move_path, chunks=chunks), parsed_lines

    raise InvalidHunkError(
        f"'{first_line}' is not a valid hunk header. Valid hunk headers: "
        f"'*** Add File: {{path}}', '*** Delete File: {{path}}', "
        f"'*** Update File: {{path}}'",
        line_number,
    )


def _parse_update_file_chunk(
    lines: list[str], line_number: int, allow_missing_context: bool
) -> tuple[UpdateFileChunk, int]:
    if not lines:
        raise InvalidHunkError("Update hunk does not contain any lines", line_number)

    if lines[0] == EMPTY_CHANGE_CONTEXT:
        change_context = None
        start_index = 1
    elif lines[0].startswith(CHANGE_CONTEXT):
        change_context = lines[0][len(CHANGE_CONTEXT) :]
        start_index = 1
    else:
        if not allow_missing_context:
            raise InvalidHunkError(
                f"Expected update hunk to start with a @@ context marker, got: '{lines[0]}'",
                line_number,
            )
        change_context = None
        start_index = 0

    if start_index >= len(lines):
        raise InvalidHunkError("Update hunk does not contain any lines", line_number + 1)

    chunk = UpdateFileChunk(change_context=change_context, old_lines=[], new_lines=[], is_end_of_file=False)
    parsed_lines = 0
    for line in lines[start_index:]:
        if line == EOF_MARKER:
            if parsed_lines == 0:
                raise InvalidHunkError("Update hunk does not contain any lines", line_number + 1)
            chunk.is_end_of_file = True
            parsed_lines += 1
            break
        if line == "":
            chunk.old_lines.append("")
            chunk.new_lines.append("")
            parsed_lines += 1
            continue
        first = line[0]
        if first == " ":
            chunk.old_lines.append(line[1:])
            chunk.new_lines.append(line[1:])
        elif first == "+":
            chunk.new_lines.append(line[1:])
        elif first == "-":
            chunk.old_lines.append(line[1:])
        else:
            if parsed_lines == 0:
                raise InvalidHunkError(
                    "Unexpected line found in update hunk: "
                    f"'{line}'. Every line should start with ' ' "
                    "(context line), '+' (added line), or '-' (removed line)",
                    line_number + 1,
                )
            break
        parsed_lines += 1

    return chunk, parsed_lines + start_index


async def apply_patch(patch: str, base_dir: Path | str | None = None) -> AffectedPaths:
    hunks = parse_patch(patch)
    base_path = Path(base_dir) if base_dir is not None else None
    return await _apply_hunks_to_files(hunks, base_path)


async def _apply_hunks_to_files(hunks: list[Hunk], base_dir: Path | None) -> AffectedPaths:
    if not hunks:
        raise PatchError("No files were modified.")

    added: list[Path] = []
    modified: list[Path] = []
    deleted: list[Path] = []

    for hunk in hunks:
        if isinstance(hunk, AddFile):
            path = _resolve_path(hunk.path, base_dir)
            parent = path.parent
            if parent and str(parent) != ".":
                await AsyncPath(parent).mkdir(parents=True, exist_ok=True)
            await AsyncPath(path).write_text(hunk.contents)
            added.append(path)
        elif isinstance(hunk, DeleteFile):
            path = _resolve_path(hunk.path, base_dir)
            await AsyncPath(path).unlink()
            deleted.append(path)
        elif isinstance(hunk, UpdateFile):
            path = _resolve_path(hunk.path, base_dir)
            new_contents = await _derive_new_contents_from_chunks(path, hunk.chunks)
            if hunk.move_path:
                dest = _resolve_path(hunk.move_path, base_dir)
                parent = dest.parent
                if parent and str(parent) != ".":
                    await AsyncPath(parent).mkdir(parents=True, exist_ok=True)
                await AsyncPath(dest).write_text(new_contents)
                await AsyncPath(path).unlink()
                modified.append(dest)
            else:
                await AsyncPath(path).write_text(new_contents)
                modified.append(path)
        else:
            raise PatchError(f"Unknown hunk type: {hunk!r}")

    return AffectedPaths(added=added, modified=modified, deleted=deleted)


def _resolve_path(path: str, base_dir: Path | None) -> Path:
    candidate = Path(path)
    if base_dir is not None and not candidate.is_absolute():
        return base_dir / candidate
    return candidate


async def _derive_new_contents_from_chunks(path: Path, chunks: list[UpdateFileChunk]) -> str:
    try:
        original_contents = await AsyncPath(path).read_text()
    except OSError as err:
        raise PatchError(f"Failed to read file to update {path}") from err

    original_lines = original_contents.split("\n")
    if original_lines and original_lines[-1] == "":
        original_lines.pop()

    replacements = _compute_replacements(original_lines, path, chunks)
    new_lines = _apply_replacements(original_lines, replacements)
    if not new_lines or new_lines[-1] != "":
        new_lines.append("")
    return "\n".join(new_lines)


def _compute_replacements(
    original_lines: list[str], path: Path, chunks: list[UpdateFileChunk]
) -> list[tuple[int, int, list[str]]]:
    replacements: list[tuple[int, int, list[str]]] = []
    line_index = 0

    for chunk in chunks:
        if chunk.change_context is not None:
            idx = _seek_sequence(original_lines, [chunk.change_context], line_index, False)
            if idx is None:
                raise PatchError(f"Failed to find context '{chunk.change_context}' in {path}")
            line_index = idx + 1

        if not chunk.old_lines:
            insertion_idx = (
                len(original_lines) - 1 if original_lines and original_lines[-1] == "" else len(original_lines)
            )
            replacements.append((insertion_idx, 0, list(chunk.new_lines)))
            continue

        pattern = list(chunk.old_lines)
        new_slice = list(chunk.new_lines)
        found = _seek_sequence(original_lines, pattern, line_index, chunk.is_end_of_file)

        if found is None and pattern and pattern[-1] == "":
            pattern = pattern[:-1]
            if new_slice and new_slice[-1] == "":
                new_slice = new_slice[:-1]
            found = _seek_sequence(original_lines, pattern, line_index, chunk.is_end_of_file)

        if found is None:
            raise PatchError(f"Failed to find expected lines in {path}:\n" + "\n".join(chunk.old_lines))

        replacements.append((found, len(pattern), new_slice))
        line_index = found + len(pattern)

    replacements.sort(key=lambda item: item[0])
    return replacements


def _apply_replacements(lines: list[str], replacements: list[tuple[int, int, list[str]]]) -> list[str]:
    out = list(lines)
    for start, old_len, new_lines in reversed(replacements):
        del out[start : start + old_len]
        for offset, line in enumerate(new_lines):
            out.insert(start + offset, line)
    return out


def _seek_sequence(lines: list[str], pattern: list[str], start: int, eof: bool) -> int | None:
    if not pattern:
        return start
    if len(pattern) > len(lines):
        return None

    search_start = len(lines) - len(pattern) if eof and len(lines) >= len(pattern) else start
    last = len(lines) - len(pattern)
    for i in range(search_start, last + 1):
        if lines[i : i + len(pattern)] == pattern:
            return i
    for i in range(search_start, last + 1):
        if _matches_with(lines, pattern, i, str.rstrip):
            return i
    for i in range(search_start, last + 1):
        if _matches_with(lines, pattern, i, str.strip):
            return i
    for i in range(search_start, last + 1):
        if _matches_with(lines, pattern, i, _normalise):
            return i
    return None


def _matches_with(lines: list[str], pattern: list[str], start: int, fn) -> bool:
    if start + len(pattern) > len(lines):
        return False
    for offset, pat in enumerate(pattern):
        if fn(lines[start + offset]) != fn(pat):
            return False
    return True


def _normalise(value: str) -> str:
    mapping = {
        "\u2010": "-",
        "\u2011": "-",
        "\u2012": "-",
        "\u2013": "-",
        "\u2014": "-",
        "\u2015": "-",
        "\u2212": "-",
        "\u2018": "'",
        "\u2019": "'",
        "\u201a": "'",
        "\u201b": "'",
        "\u201c": '"',
        "\u201d": '"',
        "\u201e": '"',
        "\u201f": '"',
        "\u00a0": " ",
        "\u2002": " ",
        "\u2003": " ",
        "\u2004": " ",
        "\u2005": " ",
        "\u2006": " ",
        "\u2007": " ",
        "\u2008": " ",
        "\u2009": " ",
        "\u200a": " ",
        "\u202f": " ",
        "\u205f": " ",
        "\u3000": " ",
    }
    return "".join(mapping.get(ch, ch) for ch in value.strip())


def format_summary(affected: AffectedPaths) -> str:
    lines = ["Success. Updated the following files:"]
    lines.extend([f"A {path}" for path in affected.added])
    lines.extend([f"M {path}" for path in affected.modified])
    lines.extend([f"D {path}" for path in affected.deleted])
    return "\n".join(lines) + "\n"


def print_summary(affected: AffectedPaths) -> None:
    sys.stdout.write(format_summary(affected))


async def main() -> None:
    args = sys.argv[1:]
    if len(args) > 1:
        sys.stderr.write("Error: apply_patch accepts exactly one argument.\n")
        sys.exit(2)
    if args:
        patch_text = args[0]
        if not isinstance(patch_text, str):
            sys.stderr.write("Error: apply_patch requires a UTF-8 PATCH argument.\n")
            sys.exit(1)
    else:
        patch_text = sys.stdin.read()
        if not patch_text:
            sys.stderr.write("Usage: apply_patch 'PATCH'\n       echo 'PATCH' | apply_patch\n")
            sys.exit(2)
    try:
        affected = await apply_patch(patch_text)
        print_summary(affected)
    except InvalidPatchError as err:
        sys.stderr.write(f"Invalid patch: {err}\n")
        sys.exit(1)
    except InvalidHunkError as err:
        sys.stderr.write(f"Invalid patch hunk on line {err.line_number}: {err}\n")
        sys.exit(1)
    except PatchError as err:
        sys.stderr.write(f"{err}\n")
        sys.exit(1)


if __name__ == "__main__":
    anyio.run(main)



import re
import sys

from exponent.core.remote_execution.types import PythonEnvInfo


def get_python_env_info() -> PythonEnvInfo:
    return PythonEnvInfo(
        interpreter_path=get_active_python_interpreter_path(),
        interpreter_version=get_active_python_interpreter_version(),
    )


def get_active_python_interpreter_path() -> str | None:
    return sys.executable


def get_active_python_interpreter_version() -> str | None:
    version = sys.version

    match = re.search(r"(\d+\.\d+\.\d+).*", version)

    if match:
        return match.group(1)

    return None
